# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# ********************************************  Keywords ***************************************************************

"""
This module provides the string keywords used throughout psyneulink

https://princetonuniversity.github.io/PsyNeuLink/Keywords.html
"""

# **********************************************************************************************************************
# ******************************************    CLASSES    *************************************************************
# **********************************************************************************************************************

# IMPLEMENTATION NOTE:
#  These classes are used for documentation purposes only.
#  The attributes of each are assigned to constants (listed in the next section of this module)
#    that are the ones actually used by the code.

__all__ = [
    'ACCUMULATOR_INTEGRATOR', 'ACCUMULATOR_INTEGRATOR_FUNCTION',
    'ADAPTIVE', 'ADAPTIVE_INTEGRATOR_FUNCTION', 'ADAPTIVE_MECHANISM', 'ADD_INPUT_PORT', 'ADD_OUTPUT_PORT',
    'ADDITIVE', 'ADDITIVE_PARAM', 'AFTER', 'ALL', 'ALLOCATION_SAMPLES', 'ANGLE',
    'ARGUMENT_THERAPY_FUNCTION', 'ARRANGEMENT', 'ASSERT', 'ASSIGN', 'ASSIGN_VALUE', 'AUTO','AUTO_ASSIGN_MATRIX',
    'AUTO_ASSOCIATIVE_PROJECTION', 'HAS_INITIALIZERS', 'AUTOASSOCIATIVE_LEARNING_MECHANISM',
    'BACKPROPAGATION_FUNCTION', 'BEFORE', 'BETA', 'BIAS', 'BOLD', 'BOTH', 'BOUNDS', 'BUFFER_FUNCTION',
    'CHANGED', 'CLAMP_INPUT', 'COMBINATION_FUNCTION_TYPE', 'COMBINE', 'COMBINE_MEANS_FUNCTION',
    'COMBINE_OUTCOME_AND_COST_FUNCTION', 'COMMAND_LINE', 'COMPARATOR_MECHANISM', 'COMPONENT', 'COMPONENT_INIT',
    'COMPONENT_PREFERENCE_SET', 'COMPOSITION', 'COMPOSITION_INTERFACE_MECHANISM', 'CONCATENATE_FUNCTION',
    'CONDITION', 'CONDITIONS', 'CONSTANT', 'ContentAddressableMemory_FUNCTION', 'CONTEXT',
    'CONTROL', 'CONTROL_MECHANISM', 'CONTROL_PROJECTION', 'CONTROL_PROJECTION_PARAMS', 'CONTROL_PROJECTIONS',
    'CONTROL_SIGNAL', 'CONTROL_SIGNAL_SPECS', 'CONTROL_SIGNALS', 'CONTROLLED_PARAMS', 'CONTROLLER', 'CORRELATION',
    'COSINE', 'COST_FUNCTION', 'COUNT', 'CROSS_ENTROPY', 'CURRENT_EXECUTION_TIME', 'CUSTOM_FUNCTION', 'CYCLE',
    'DDM_MECHANISM', 'DECAY', 'DEFAULT', 'DEFAULT_CONTROL_MECHANISM', 'DEFAULT_MATRIX',
    'DEFAULT_PREFERENCE_SET_OWNER', 'DEFAULT_PROCESSING_MECHANISM', 'DEFAULT_SYSTEM', 'DEFAULT_VARIABLE',
    'DEFERRED_ASSIGNMENT', 'DEFERRED_DEFAULT_NAME', 'DEFERRED_INITIALIZATION',
    'DIFFERENCE', 'DIFFERENCE', 'DIFFUSION', 'DISABLE', 'DISABLE_PARAM', 'DIST_FUNCTION_TYPE', 'DIST_MEAN',
    'DIST_SHAPE', 'DISTANCE_FUNCTION', 'DISTANCE_METRICS', 'DISTRIBUTION_FUNCTION_TYPE', 'DIVISION',
    'DRIFT_DIFFUSION_INTEGRATOR_FUNCTION', 'DUAL_ADAPTIVE_INTEGRATOR_FUNCTION',
    'EID_SIMULATION', 'EID_FROZEN', 'ENABLE_CONTROLLER', 'ENABLED', 'ENERGY', 'ENTROPY', 'ERROR_DERIVATIVE_FUNCTION',
    'EUCLIDEAN', 'EVC_MECHANISM', 'EVC_SIMULATION', 'EXAMPLE_FUNCTION_TYPE',
    'EXECUTING', 'EXECUTION', 'EXECUTION_ID', 'EXECUTION_PHASE',
    'EXPONENTIAL', 'EXPONENT', 'EXPONENTIAL_DIST_FUNCTION', 'EXPONENTIAL_FUNCTION', 'EXPONENTS',
    'FITZHUGHNAGUMO_INTEGRATOR_FUNCTION', 'FINAL', 'FLAGS', 'FULL', 'FULL_CONNECTIVITY_MATRIX',
    'FUNCTION', 'FUNCTIONS', 'FUNCTION_COMPONENT_CATEGORY','FUNCTION_CHECK_ARGS',
    'FUNCTION_OUTPUT_TYPE', 'FUNCTION_OUTPUT_TYPE_CONVERSION', 'FUNCTION_PARAMS',
    'GAIN', 'GAMMA_DIST_FUNCTION', 'GATE', 'GATING', 'GATING_MECHANISM', 'GATING_ALLOCATION', 'GATING_PROJECTION',
    'GATING_PROJECTION_PARAMS', 'GATING_PROJECTIONS', 'GATING_SIGNAL', 'GATING_SIGNAL_SPECS', 'GATING_SIGNALS',
    'GAUSSIAN', 'GAUSSIAN_FUNCTION', 'GILZENRAT_INTEGRATOR_FUNCTION',
    'GRADIENT_OPTIMIZATION_FUNCTION', 'GRID_SEARCH_FUNCTION',
    'HARD_CLAMP', 'HEBBIAN_FUNCTION', 'HETERO', 'HIGH', 'HOLLOW_MATRIX', 'IDENTITY_MATRIX', 'INCREMENT', 'INDEX',
    'INIT_EXECUTE_METHOD_ONLY', 'INIT_FULL_EXECUTE_METHOD', 'INIT_FUNCTION_METHOD_ONLY', 'INITIAL_VALUES',
    'INITIALIZE_CYCLE', 'INITIALIZATION', 'INITIALIZED', 'INITIALIZER', 'INITIALIZING', 'INITIALIZATION_STATUS',
    'INPUT', 'INPUT_LABELS_DICT', 'INPUT_PORT', 'INPUT_PORTS', 'INPUT_PORT_PARAMS', 'INPUT_PORT_VARIABLES',
    'INPUTS_DIM', 'INSTANTANEOUS_MODE_VALUE', 'INTEGRATION_TYPE', 'INTEGRATOR_FUNCTION', 'INTEGRATOR_FUNCTION',
    'INTEGRATOR_FUNCTION_TYPE', 'INTEGRATOR_MECHANISM', 'INTEGRATOR_MODE_VALUE', 'INTERCEPT', 'INTERNAL',
    'INTERNAL_ONLY', 'K_VALUE', 'KOHONEN_FUNCTION', 'KOHONEN_MECHANISM', 'KOHONEN_LEARNING_MECHANISM', 'KWTA_MECHANISM',
    'LABELS', 'LCA_MECHANISM', 'LEAKY_COMPETING_INTEGRATOR_FUNCTION', 'LEAK', 'LEARNING', 'LEARNED_PARAM',
    'LEARNED_PROJECTION', 'LEARNING_FUNCTION_TYPE', 'LEARNING_MECHANISM', 'LEARNING_PROJECTION',
    'LEARNING_PROJECTION_PARAMS', 'LEARNING_RATE', 'LEARNING_SIGNAL', 'LEARNING_SIGNAL_SPECS', 'LEARNING_SIGNALS',
    'LINEAR', 'LINEAR_COMBINATION_FUNCTION', 'LINEAR_FUNCTION', 'LINEAR_MATRIX_FUNCTION', 'LOG_ENTRIES',
    'LOGISTIC_FUNCTION', 'LOW', 'LVOC_CONTROL_MECHANISM', 'L0', 'L1',
    'MAKE_DEFAULT_GATING_MECHANISM', 'MAPPING_PROJECTION', 'MAPPING_PROJECTION_PARAMS', 'MASKED_MAPPING_PROJECTION',
    'MATRIX', 'MATRIX_KEYWORD_NAMES', 'MATRIX_KEYWORD_SET', 'MATRIX_KEYWORD_VALUES', 'MATRIX_KEYWORDS','MatrixKeywords',
    'MAX_ABS_VAL', 'MAX_ABS_INDICATOR', 'MAX_ABS_DIFF', 'MAX_INDICATOR', 'MAX_VAL', 'MAYBE',
    'MECHANISM', 'MECHANISM_COMPONENT_CATEGORY', 'MECHANISM_DEFAULT', 'MECHANISM_DEFAULTInputValue',
    'MECHANISM_DEFAULTParams', 'MECHANISM_EXECUTED_LOG_ENTRY', 'MECHANISM_NAME', 'MECHANISM_PARAM_VALUE',
    'MECHANISM_TYPE', 'MECHANISM_VALUE',
    'METRIC', 'MIN_VAL', 'MODE', 'MODULATES','MODULATION', 'MODULATORY_PROJECTION', 'MODULATORY_SIGNAL',
    'MODULATORY_SIGNALS', 'MONITOR', 'MONITOR_FOR_CONTROL', 'MONITOR_FOR_LEARNING', 'MONITOR_FOR_MODULATION',
    'MODEL_SPEC_ID_GENERIC', 'MODEL_SPEC_ID_INPUT_PORTS', 'MODEL_SPEC_ID_OUTPUT_PORTS', 'MODEL_SPEC_ID_PSYNEULINK',
    'MODEL_SPEC_ID_SENDER_MECH', 'MODEL_SPEC_ID_SENDER_PORT', 'MODEL_SPEC_ID_RECEIVER_MECH',
    'MODEL_SPEC_ID_RECEIVER_PORT',
    'MODEL_SPEC_ID_PARAMETER_SOURCE', 'MODEL_SPEC_ID_PARAMETER_VALUE', 'MODEL_SPEC_ID_TYPE', 'MSE',
    'MULTIPLICATIVE', 'MULTIPLICATIVE_PARAM', 'MUTUAL_ENTROPY',
    'NAME', 'NEWEST',  'NODE', 'NodeRoles', 'NOISE', 'NORMAL_DIST_FUNCTION', 'NORMED_L0_SIMILARITY',
    'OBJECTIVE_FUNCTION_TYPE', 'OBJECTIVE_MECHANISM', 'OBJECTIVE_MECHANISM_OBJECT', 'OFF', 'OFFSET', 'OLDEST',
    'ON',  'ONLINE', 'OPERATION', 'OPTIMIZATION_FUNCTION_TYPE', 'ORIGIN','ORNSTEIN_UHLENBECK_INTEGRATOR_FUNCTION',
    'OUTCOME', 'OUTCOME_FUNCTION', 'OUTPUT', 'OUTPUT_LABELS_DICT', 'OUTPUT_MEAN', 'OUTPUT_MEDIAN', 'OUTPUT_STD_DEV',
    'OUTPUT_PORT', 'OUTPUT_PORT_PARAMS', 'output_port_spec_to_parameter_name', 'OUTPUT_PORTS', 'OUTPUT_TYPE',
    'OUTPUT_VARIANCE', 'OVERRIDE', 'OVERRIDE_PARAM', 'OVERWRITE',
    'OWNER', 'OWNER_EXECUTION_COUNT', 'OWNER_EXECUTION_TIME', 'OWNER_VALUE', 'OWNER_VARIABLE',
    'PARAM_CLASS_DEFAULTS', 'PARAM_INSTANCE_DEFAULTS', 'PARAMETER_PORT', 'PARAMETER_PORT_PARAMS',
    'PARAMETER_PORTS', 'PARAMS', 'PARAMS_DICT', 'PARAMS_CURRENT', 'PATHWAY', 'PATHWAY_PROJECTION', 'PEARSON',
    'PREDICTION_MECHANISM', 'PREDICTION_MECHANISMS', 'PREDICTION_MECHANISM_OUTPUT', 'PREDICTION_MECHANISM_PARAMS',
    'PREDICTION_MECHANISM_TYPE', 'PREFS_ARG', 'PREF_BASE_VALUE', 'PREF_CURRENT_VALUE', 'PREFERENCE_SET',
    'PREFERENCE_SET_NAME', 'PREF_LEVEL', 'PREFS', 'PREFS_OWNER', 'PREVIOUS_VALUE', 'PRIMARY', 'PROB', 'PROB_INDICATOR',
    'PROCESS', 'PROCESS_COMPONENT_CATEGORY', 'PROCESS_DEFAULT_MECHANISM', 'PROCESS_DEFAULT_PROJECTION_FUNCTION',
    'PROCESS_EXECUTE', 'PROCESS_INIT', 'PROCESSES', 'PROCESSES_DIM', 'PROCESSING', 'PROCESSING_MECHANISM',
    'PRODUCT', 'PROGRESS_BAR_CHAR', 'PROJECTION', 'PROJECTION_DIRECTION', 'PROJECTION_PARAMS', 'PROJECTION_SENDER',
    'PROJECTION_TYPE', 'PROJECTIONS', 'PROJECTION_COMPONENT_CATEGORY', 'QUOTIENT',
    'RANDOM', 'RANDOM_CONNECTIVITY_MATRIX', 'RATE', 'RATIO', 'REARRANGE_FUNCTION', 'RECEIVER', 'RECEIVER_ARG',
    'RECURRENT_TRANSFER_MECHANISM', 'REDUCE_FUNCTION', 'REFERENCE_VALUE', 'REINITIALIZE', 'REINITIALIZE_WHEN',
    'RELU_FUNCTION', 'REST', 'RESULT', 'RESULTS', 'ROLES', 'RL_FUNCTION', 'RUN',
    'SAMPLE', 'SAVE_ALL_VALUES_AND_POLICIES', 'SCALAR', 'SCALE', 'SCHEDULER', 'SELF', 'SENDER',
    'SEPARATOR_BAR', 'SIMPLE', 'SIMPLE_INTEGRATOR_FUNCTION', 'SINGLETON', 'SIZE', 'SLOPE', 'SOFT_CLAMP',
    'SOFTMAX_FUNCTION', 'SOURCE', 'SSE', 'STABILITY_FUNCTION', 'STANDARD_ARGS', 'STANDARD_DEVIATION', 'STANDARD_OUTPUT_PORTS',
    'PORT', 'PORT_COMPONENT_CATEGORY', 'PORT_CONTEXT', 'Port_Name', 'Port_ParamS', 'PORT_PREFS', 'PORT_TYPE',
    'port_value', 'PORTS', 'SUBTRACTION', 'SUM',
    'SYSTEM', 'SYSTEM_COMPONENT_CATEGORY', 'SYSTEM_DEFAULT_CONTROLLER', 'SYSTEM_INIT',
    'TARGET', 'TARGET_MECHANISM', 'TARGET_LABELS_DICT', 'TERMINAL', 'THRESHOLD', 'TIME', 'TIME_STEP_SIZE',
    'TIME_STEPS_DIM', 'TRAINING_SET', 'TRANSFER_FUNCTION_TYPE', 'TRANSFER_MECHANISM', 'TRANSFER_WITH_COSTS_FUNCTION',
    'TRIAL', 'TRIALS_DIM',
    'UNCHANGED', 'UNIFORM_DIST_FUNCTION', 'USER_DEFINED_FUNCTION', 'USER_DEFINED_FUNCTION_TYPE', 'USER_PARAMS',
    'VALUES', 'VALIDATE', 'VALIDATION', 'VALUE', 'VALUE_ASSIGNMENT', 'VALUE_FUNCTION', 'VARIABLE', 'VARIANCE',
    'VECTOR', 'WALD_DIST_FUNCTION', 'WEIGHT', 'WEIGHTS', 'X_0',
]


class NodeRoles:
    """
    Attributes
    ----------

    ORIGIN
        A `ProcessingMechanism <ProcessingMechanism>` that is the first Mechanism of a `Process` and/or `System`,
        and that receives the input to the Process or System when it is :ref:`executed or run <Run>`.  A Process may
        have only one `ORIGIN` Mechanism, but a System may have many.  Note that the `ORIGIN`
        Mechanism of a Process is not necessarily an `ORIGIN` of the System to which it belongs, as it may receive
        `Projections <Projection>` from other Processes in the System. The `ORIGIN` Mechanisms of a Process or
        System are listed in its :keyword:`origin_mechanisms` attribute, and can be displayed using its :keyword:`show`
        method.  For additional details about `ORIGIN` Mechanisms in Processes, see
        `Process Mechanisms <Process_Mechanisms>` and `Process Input and Output <Process_Input_And_Output>`;
        and for Systems see `System Mechanisms <System_Mechanisms>` and
        `System Input and Initialization <System_Execution_Input_And_Initialization>`.

    INTERNAL
        A `ProcessingMechanism <ProcessingMechanism>` that is not designated as having any other status.

    CYCLE
        A `ProcessingMechanism <ProcessingMechanism>` that is *not* an `ORIGIN` Mechanism, and receives a `Projection
        <Projection>` that closes a recurrent loop in a `Process` and/or `System`.  If it is an `ORIGIN` Mechanism, then
        it is simply designated as such (since it will be assigned input and therefore be initialized in any event).

    INITIALIZE_CYCLE
        A `ProcessingMechanism <ProcessingMechanism>` that is the `sender <Projection_Base.sender>` of a
        `Projection <Projection>` that closes a loop in a `Process` or `System`, and that is not an `ORIGIN` Mechanism
        (since in that case it will be initialized in any event). An `initial value  <Run_InitialValues>` can be
        assigned to such Mechanisms, that will be used to initialize the Process or System when it is first run.  For
        additional information, see `Run <Run_Initial_Values>`, `System Mechanisms <System_Mechanisms>` and
        `System Input and Initialization <System_Execution_Input_And_Initialization>`.

    TERMINAL
        A `ProcessingMechanism <ProcessingMechanism>` that is the last Mechanism of a `Process` and/or `System`, and
        that provides the output to the Process or System when it is `executed or run <Run>`.  A Process may
        have only one `TERMINAL` Mechanism, but a System may have many.  Note that the `TERMINAL`
        Mechanism of a process is not necessarily a `TERMINAL` Mechanism of the System to which it belongs,
        as it may send projections to other processes in the System (see `example
        <LearningProjection_Output_vs_Terminal_Figure>`).  The `TERMINAL` Mechanisms of a Process or System are listed in
        its :keyword:`terminalMechanisms` attribute, and can be displayed using its :keyword:`show` method.  For
        additional details about `TERMINAL` Mechanisms in Processes, see `Process_Mechanisms` and
        `Process_Input_And_Output`; and for Systems see `System_Mechanisms`.

    SINGLETON
        A `ProcessingMechanism <ProcessingMechanism>` that is the only Mechanism in a `Process` and/or `System`.
        It can serve the functions of an `ORIGIN` and/or a `TERMINAL` Mechanism.

    LEARNING
        A `LearningMechanism <LearningMechanism>` in a `Process` and/or `System`.

    TARGET
        A `ComparatorMechanism` of a `Process` and/or `System` configured for learning that receives a target value
        from its `execute <ComparatorMechanism.ComparatorMechanism.execute>` or
        `run <ComparatorMechanism.ComparatorMechanism.execute>` method.  It is usually (but not necessarily)
        associated with the `TERMINAL` Mechanism of the Process or System. The `TARGET` Mechanisms of a Process or
        System are listed in its :keyword:`target_nodes` attribute, and can be displayed using its
        :keyword:`show` method.  For additional details, see `TARGET Mechanisms <LearningMechanism_Targets>`,
        `learning sequence <Process_Learning_Sequence>`, and specifying `target values <Run_Targets>`.


    """
    def __init__(self):
        self.ORIGIN = ORIGIN
        self.INTERNAL = INTERNAL
        self.CYCLE = CYCLE
        self.INITIALIZE_CYCLE = INITIALIZE_CYCLE
        self.TERMINAL = TERMINAL
        self.SINGLETON = SINGLETON
        self.LEARNING = LEARNING
        self.TARGET = TARGET

NODE = 'NODE'

class MatrixKeywords:
    """
    Attributes
    ----------

    IDENTITY_MATRIX
        a square matrix of 1's along the diagonal, 0's elsewhere; this requires that the length of the sender and
        receiver values are the same.

    HOLLOW_MATRIX
        a square matrix of 0's along the diagonal, 1's elsewhere; this requires that the length of the sender and
        receiver values are the same.

    FULL_CONNECTIVITY_MATRIX
        a matrix that has a number of rows equal to the length of the sender's value, and a number of columns equal
        to the length of the receiver's value, all the elements of which are 1's.

    RANDOM_CONNECTIVITY_MATRIX
        a matrix that has a number of rows equal to the length of the sender's value, and a number of columns equal
        to the length of the receiver's value, all the elements of which are filled with random values uniformly
        distributed between 0 and 1.

    AUTO_ASSIGN_MATRIX
        if the sender and receiver are of equal length, an `IDENTITY_MATRIX` is assigned;  otherwise, a
        `FULL_CONNECTIVITY_MATRIX` is assigned.

    DEFAULT_MATRIX
        used if no matrix specification is provided in the constructor;  it presently assigns an `IDENTITY_MATRIX`.

    """
    def __init__(self):
        # self.MATRIX = MATRIX
        self.IDENTITY_MATRIX = IDENTITY_MATRIX
        self.HOLLOW_MATRIX = HOLLOW_MATRIX
        self.INVERSE_HOLLOW_MATRIX = INVERSE_HOLLOW_MATRIX
        self.FULL_CONNECTIVITY_MATRIX = FULL_CONNECTIVITY_MATRIX
        self.RANDOM_CONNECTIVITY_MATRIX = RANDOM_CONNECTIVITY_MATRIX
        self.AUTO_ASSIGN_MATRIX = AUTO_ASSIGN_MATRIX
        self.DEFAULT_MATRIX = DEFAULT_MATRIX

    def _values(self):
        return list(self.__dict__.values())

    def _set(self):
        return set(self.__dict__.values())

    def _names(self):
        return list(self.__dict__)


MATRIX = "matrix"
IDENTITY_MATRIX = "IdentityMatrix"
HOLLOW_MATRIX = "HollowMatrix"
INVERSE_HOLLOW_MATRIX = "InverseHollowMatrix"
FULL_CONNECTIVITY_MATRIX = "FullConnectivityMatrix"
RANDOM_CONNECTIVITY_MATRIX = "RandomConnectivityMatrix"
AUTO_ASSIGN_MATRIX = 'AutoAssignMatrix'
DEFAULT_MATRIX = AUTO_ASSIGN_MATRIX
# DEFAULT_MATRIX = IDENTITY_MATRIX

MATRIX_KEYWORDS = MatrixKeywords()
MATRIX_KEYWORD_SET = MATRIX_KEYWORDS._set()
MATRIX_KEYWORD_VALUES = MATRIX_KEYWORDS._values()
MATRIX_KEYWORD_NAMES = MATRIX_KEYWORDS._names()


class DistanceMetrics:
    """Distance between two arrays.

    Each keyword specifies a metric for the distance between two arrays, :math:`a_1` and :math:`a_2`, of equal length
    for which *len* is their length, :math:`\\bar{a}` is the mean of an array, :math:`\\sigma_{a}` the standard
    deviation of an array, and :math:`w_{a_1a_2}` a coupling coefficient ("weight") between a pair of elements,
    one from each array:

    Attributes
    ----------

    MAX_ABS_DIFF
        :math:`d = \\max(|a_1-a_2|)`

    DIFFERENCE
        (can also be referenced as L0)\n
        :math:`d = \\sum\\limits^{len}(|a_1-a_2|)`

    EUCLIDEAN
        (can also be referenced as L1)\n
        :math:`d = \\sum\\limits^{len}\\sqrt{(a_1-a_2)^2}`

    COSINE
        :math:`d = 1 - \\frac{\\sum\\limits^{len}a_1a_2}{\\sqrt{\\sum\\limits^{len}a_1^2}\\sqrt{\\sum\\limits^{len}a_2^2}}`

    CORRELATION
        :math:`d = 1 - \\left|\\frac{\\sum\\limits^{len}(a_1-\\bar{a}_1)(a_2-\\bar{a}_2)}{(len-1)\\sigma_{a_1}\\sigma_{
        a_2}}\\right|`

    COMMENT:
    PEARSON
        <Description>
    COMMENT

    ENTROPY and CROSS_ENTROPY
        :math:`d = \\sum\\limits^{len}a_1log(a_2)`

    ENERGY:
        :math:`d = -\\frac{1}{2}\\sum\\limits_{i,j}a_{1_i}a_{2_j}w_{ij}`

    """
    def __init__(self):
        self.MAX_ABS_DIFF = MAX_ABS_DIFF
        self.DIFFERENCE = DIFFERENCE
        self.L0 = L0
        self.NORMED_L0_SIMILARITY = NORMED_L0_SIMILARITY
        self.EUCLIDEAN = EUCLIDEAN
        self.L1 = L1
        self.ANGLE = ANGLE
        self.CORRELATION = CORRELATION
        # self.PEARSON = PEARSON
        self.COSINE = COSINE
        self.ENTROPY = CROSS_ENTROPY
        self.CROSS_ENTROPY = CROSS_ENTROPY
        self.ENERGY = ENERGY

    def _values(self):
        return list(self.__dict__.values())

    def _set(self):
        return set(self.__dict__.values())

    def _names(self):
        return list(self.__dict__)

    def _is_metric(metric):
        if metric in DISTANCE_METRICS_SET:
            return True
        else:
            return False


METRIC = 'metric'
DIFFERENCE = 'difference'
L0 = DIFFERENCE
NORMED_L0_SIMILARITY = 'normed_L0_similarity'
MAX_ABS_DIFF = 'max_abs_diff'
EUCLIDEAN = 'euclidean'
L1 = EUCLIDEAN
ANGLE = 'angle'
CORRELATION = 'correlation'
COSINE = 'cosine'
PEARSON = 'Pearson'
ENTROPY = 'cross-entropy'
CROSS_ENTROPY = 'cross-entropy'
ENERGY = 'energy'

DISTANCE_METRICS = DistanceMetrics()
DISTANCE_METRICS_SET = DISTANCE_METRICS._set()
DISTANCE_METRICS_VALUES = DISTANCE_METRICS._values()
DISTANCE_METRICS_NAMES = DISTANCE_METRICS._names()

ENERGY = 'energy'
ENTROPY = 'entropy'


# **********************************************************************************************************************
# ******************************************    CONSTANTS  *************************************************************
# **********************************************************************************************************************

ON = True
OFF = False
DEFAULT = False
# AUTO = True  # MODIFIED 7/14/17 CW
ASSERT = True

# Used by initDirective
INIT_FULL_EXECUTE_METHOD = 'init using the full base class execute method'
INIT_EXECUTE_METHOD_ONLY = 'init using only the subclass _execute method'
INIT_FUNCTION_METHOD_ONLY = 'init using only the subclass __function__ method'


#region ---------------------------------------------    GENERAL    ----------------------------------------------------
# General

ALL = 'all'
BOTH = 'both'
MAYBE = 0.5

SEPARATOR_BAR = ' | '
PROGRESS_BAR_CHAR = '.'
# VALUE_SUFFIX = '_value'
SELF = 'self'
FLAGS = 'flags'
INITIALIZATION_STATUS = 'initialization_status'
EXECUTION_PHASE = 'execution_phase'
SOURCE = 'source'
INITIALIZING = " INITIALIZING "  # Used as status and context for Log
INITIALIZED = " INITIALIZED "  # Used as status
EXECUTING = " EXECUTING " # Used in context for Log and ReportOutput pref
ASSIGN_VALUE = ': Assign value'
VALIDATE = 'Validate'
COMMAND_LINE = "COMMAND_LINE"
CHANGED = 'CHANGED'
UNCHANGED = 'UNCHANGED'
ENABLED = 'ENABLED'
STATEFUL_ATTRIBUTES = 'stateful_attributes'
COUNT = 'COUNT'
BOLD = 'bold'
BEFORE = 'before'
AFTER = 'after'
ONLINE = 'online'
INPUT = 'input'
OUTPUT = 'output'
RANDOM = 'random'
OLDEST = 'oldest'
NEWEST = 'newest'

EID_SIMULATION = '-sim'
EID_FROZEN = '-frozen'

#endregion

#region -------------------------------------------    Preferences    --------------------------------------------------

PREFS = "Prefs"
PREFS_OWNER = "PrefsOwner"
PREF_LEVEL = 'PrefLevel'
PREF_CURRENT_VALUE = 'PrefCurrentValue'
PREF_BASE_VALUE = 'PrefBaseValue'
PREFERENCE_SET_NAME = 'PreferenceSetName'
DEFAULT_PREFERENCE_SET_OWNER = 'DefaultPreferenceSetOwner'
#endregion

#region --------------------------------------------    PREFERENCES    -------------------------------------------------

PREFERENCE_SET = 'PreferenceSet'
COMPONENT_PREFERENCE_SET = 'BasePreferenceSet'
# COMPONENT_PREFERENCE_SET = 'PreferenceSet'
#endregion

#region ------------------------------------------------   LOG    ------------------------------------------------------

TIME = 'time'
LOG_ENTRIES = 'LOG_ENTRIES'
INITIALIZATION = 'INITIALIZATION'
VALIDATION = 'VALIDATION'
EXECUTION = 'EXECUTION'
PROCESSING = 'PROCESSING'
VALUE_ASSIGNMENT = 'VALUE_ASSIGNMENT'
FINAL = 'FINAL'


#endregion

#region ----------------------------------------------   COMPONENT   ---------------------------------------------------

COMPOSITION = 'COMPOSITION'

#endregion

#region ----------------------------------------------   COMPONENT   ---------------------------------------------------

COMPONENT = 'COMPONENT'

# Standard arg / attribute names:
VARIABLE = "variable"
DEFAULT_VARIABLE = "default_variable"
VALUE = "value"
PREVIOUS_VALUE = 'previous_value'
LABELS = 'labels'
PARAMS = "params"
NAME = "name"
PREFS_ARG = "prefs"
CONTEXT = "context"
STANDARD_ARGS = {NAME, VARIABLE, VALUE, PARAMS, PREFS_ARG, CONTEXT}

INITIAL_VALUES = 'initial_values'
CURRENT_EXECUTION_TIME = 'current_execution_time'
EXECUTION_ID = 'execution_id'

# inputs list/ndarray:
TRIALS_DIM = 0
TIME_STEPS_DIM = 1
PROCESSES_DIM = 2
INPUTS_DIM = 3

COMPONENT_INIT = 'Component.__init__'
DEFERRED_INITIALIZATION = 'Deferred Init'
DEFERRED_ASSIGNMENT = 'Deferred Assignment'
DEFERRED_DEFAULT_NAME = 'DEFERRED_DEFAULT_NAME'
USER_PARAMS = 'user_params' # Parameters available to user for inspection in user_params dict
FUNCTION = "function" # Parameter name for function, method, or type to instantiate and assign to self.execute
FUNCTION_PARAMS  = "function_params" # Parameters used to instantiate or assign to a FUNCTION

PARAM_CLASS_DEFAULTS = "paramClassDefaults"        # "Factory" default params for a Function
PARAM_INSTANCE_DEFAULTS = "paramInstanceDefaults" # Parameters used to instantiate a Function; supercede paramClassDefaults
PARAMS_CURRENT = "paramsCurrent"                  # Parameters currently in effect for an instance of a Function
                                                   #    in general, this includes params specifed as arg in a
                                                   #    to Function.execute;  however, there are some exceptions
                                                   #    in which those are kept separate from paramsCurrent (see DDM)
FUNCTION_CHECK_ARGS = 'super._check_args' # Use for "context" arg
FUNCTION_OUTPUT_TYPE_CONVERSION = "enable_output_type_conversion"  # Used in Function Components to set output type

#endregion

#region ----------------------------------------    COMPONENT SUBCLASSES  ----------------------------------------------

# Component Categories   -----------------

SYSTEM_COMPONENT_CATEGORY = "System"
PROCESS_COMPONENT_CATEGORY = "Process"
MECHANISM_COMPONENT_CATEGORY = "Mechanism_Base"
PORT_COMPONENT_CATEGORY = "Port_Base"
PROJECTION_COMPONENT_CATEGORY = "Projection_Base"
FUNCTION_COMPONENT_CATEGORY = "Function_Base"

# Component TYPES  -----------------

# Mechanisms:
PROCESSING_MECHANISM = "ProcessingMechanism"
ADAPTIVE_MECHANISM = "ModulatoryMechanism"
LEARNING_MECHANISM = "LearningMechanism"
CONTROL_MECHANISM = "ControlMechanism"
TARGET_MECHANISM = "TargetMechanism"
GATING_MECHANISM = 'GatingMechanism'
AUTOASSOCIATIVE_LEARNING_MECHANISM = 'AutoAssociativeLearningMechanism'
KOHONEN_LEARNING_MECHANISM = 'KohonenLearningMechanism'

# Ports:
INPUT_PORT = "InputPort"
PROCESS_INPUT_PORT = "ProcessInputPort"
SYSTEM_INPUT_PORT = "SystemInputPort"
PARAMETER_PORT = "ParameterPort"
OUTPUT_PORT = "OutputPort"
MODULATORY_SIGNAL = 'ModulatorySignal'
LEARNING_SIGNAL = 'LearningSignal'
CONTROL_SIGNAL = 'ControlSignal'
GATING_SIGNAL = 'GatingSignal'

# Projections:
MAPPING_PROJECTION = "MappingProjection"
AUTO_ASSOCIATIVE_PROJECTION = "AutoAssociativeProjection"
MASKED_MAPPING_PROJECTION = 'MaskedMappingProjection'
LEARNING_PROJECTION = "LearningProjection"
CONTROL_PROJECTION = "ControlProjection"
GATING_PROJECTION = "GatingProjection"
PATHWAY_PROJECTION = "PathwayProjection"
PATHWAY_PROJECTIONS = "PathwayProjections"
MODULATORY_PROJECTION = "ModulatoryProjection"
MODULATORY_PROJECTIONS = "ModulatoryProjections"


# Function:
EXAMPLE_FUNCTION_TYPE = "EXAMPLE FUNCTION"
USER_DEFINED_FUNCTION_TYPE = "USER DEFINED FUNCTION TYPE"
COMBINATION_FUNCTION_TYPE = "COMBINATION FUNCTION TYPE"
DIST_FUNCTION_TYPE = "DIST FUNCTION TYPE"
STATEFUL_FUNCTION_TYPE = "STATEFUL FUNCTION TYPE"
MEMORY_FUNCTION_TYPE = "MEMORY FUNCTION TYPE"
INTEGRATOR_FUNCTION_TYPE = "INTEGRATOR FUNCTION TYPE"
TRANSFER_FUNCTION_TYPE = "TRANSFER FUNCTION TYPE"
LEABRA_FUNCTION_TYPE = "LEABRA FUNCTION TYPE"
DISTRIBUTION_FUNCTION_TYPE = "DISTRIBUTION FUNCTION TYPE"
OBJECTIVE_FUNCTION_TYPE = "OBJECTIVE FUNCTION TYPE"
OPTIMIZATION_FUNCTION_TYPE = "OPTIMIZATION FUNCTION TYPE"
LEARNING_FUNCTION_TYPE = 'LEARNING FUNCTION TYPE'
NORMALIZING_FUNCTION_TYPE = "NORMALIZING FUNCTION TYPE"
INTERFACE_FUNCTION_TYPE = "INTERFACE FUNCTION TYPE"
SELECTION_FUNCTION_TYPE = "SELECTION FUNCTION TYPE"


# Component SUBTYPES -----------------

# ControlMechanism:
DEFAULT_CONTROL_MECHANISM = "DefaultControlMechanism"
OPTIMIZATION_CONTROL_MECHANISM = 'OptimizationControlMechanism'
EVC_MECHANISM = "EVCControlMechanism"
LVOC_CONTROL_MECHANISM = 'LVOCControlMechanism'

# ObjectiveMechanisms:
OBJECTIVE_MECHANISM_OBJECT = "ObjectiveMechanism"
COMPARATOR_MECHANISM = "ComparatorMechanism"
PREDICTION_ERROR_MECHANISM = "PredictionErrorMechanism"

# ProcessingMechanisms:
TRANSFER_MECHANISM = "TransferMechanism"
LEABRA_MECHANISM = "LeabraMechanism"
RECURRENT_TRANSFER_MECHANISM = "RecurrentTransferMechanism"
CONTRASTIVE_HEBBIAN_MECHANISM = "ContrastiveHebbianMechanism"
LCA_MECHANISM = "LCAMechanism"
KOHONEN_MECHANISM = 'KohonenMechanism'
KWTA_MECHANISM = "KWTAMechanism"
INTEGRATOR_MECHANISM = "IntegratorMechanism"
DDM_MECHANISM = "DDM"
COMPOSITION_INTERFACE_MECHANISM = "CompositionInterfaceMechanism"
PROCESSING_MECHANISM = "ProcessingMechanism"

# Functions:
ARGUMENT_THERAPY_FUNCTION = "Contradiction Function"
USER_DEFINED_FUNCTION = "USER DEFINED FUNCTION"

# CombinationFunctions:
REDUCE_FUNCTION = "Reduce Function"
CONCATENATE_FUNCTION = "Concatenate Function"
REARRANGE_FUNCTION = 'Rearrange Function'
LINEAR_COMBINATION_FUNCTION = "LinearCombination Function"
COMBINE_MEANS_FUNCTION = "CombineMeans Function"

# TransferFunctions:
IDENTITY_FUNCTION = 'Identity Function'
LINEAR_FUNCTION = "Linear Function"
LEABRA_FUNCTION = "Leabra Function"
EXPONENTIAL_FUNCTION = "Exponential Function"
LOGISTIC_FUNCTION = "Logistic Function"
RELU_FUNCTION = "ReLU Function"
GAUSSIAN_FUNCTION = "Gaussian Function"
GAUSSIAN_DISTORT_FUNCTION = "GaussianDistort Function"
SOFTMAX_FUNCTION = 'SoftMax Function'
LINEAR_MATRIX_FUNCTION = "LinearMatrix Function"
TRANSFER_WITH_COSTS_FUNCTION = "TransferWithCosts Function"

# SelectionFunctions:
ONE_HOT_FUNCTION = "OneHot Function"

# IntegratorFunctions:
INTEGRATOR_FUNCTION = "IntegratorFunction Function"
SIMPLE_INTEGRATOR_FUNCTION = "SimpleIntegrator Function"
INTERACTIVE_ACTIVATION_INTEGRATOR_FUNCTION = "Interactive Activation IntegratorFunction Function"
ACCUMULATOR_INTEGRATOR_FUNCTION = "AccumulatorIntegrator Function"
FITZHUGHNAGUMO_INTEGRATOR_FUNCTION = "FitzHughNagumoIntegrator Function"
DUAL_ADAPTIVE_INTEGRATOR_FUNCTION = "DualAdaptiveIntegrator Function"
ACCUMULATOR_INTEGRATOR = "AccumulatorIntegrator"  # (7/19/17 CW) added for MappingProjection.py
LEAKY_COMPETING_INTEGRATOR_FUNCTION = 'LeakyCompetingIntegrator Function'
ADAPTIVE_INTEGRATOR_FUNCTION = "AdaptiveIntegrator Function"
GILZENRAT_INTEGRATOR_FUNCTION = "GilzenratDecisionIntegrator Function"
DRIFT_DIFFUSION_INTEGRATOR_FUNCTION = "DriftDiffusionIntegrator Function"
ORNSTEIN_UHLENBECK_INTEGRATOR_FUNCTION = "OU IntegratorFunction Function"

# MemoryFunctions:
BUFFER_FUNCTION = 'Buffer Function'
ContentAddressableMemory_FUNCTION = 'ContentAddressableMemory Function'

# OptimizationFunctions:
GRADIENT_OPTIMIZATION_FUNCTION = "GradientOptimization Function"
GRID_SEARCH_FUNCTION = 'GridSearch Function'

# LearningFunctions:
GAUSSIAN_PROCESS_FUNCTION = 'GaussianProcess Function'
HEBBIAN_FUNCTION = 'Hebbian Learning Function'
CONTRASTIVE_HEBBIAN_FUNCTION = 'ContrastiveHebbian Learning Function'
KOHONEN_FUNCTION = 'Kohonen Learning Function'
RL_FUNCTION = 'ReinforcementLearning Function'
BACKPROPAGATION_FUNCTION = 'Backpropagation Learning Function'
TDLEARNING_FUNCTION = "TD Learning Function"
PREDICTION_ERROR_DELTA_FUNCTION = "PredictionErrorDelta Function"
ERROR_DERIVATIVE_FUNCTION = 'Error Derivative Function'

# Distributionfunctions
NORMAL_DIST_FUNCTION = "Normal Distribution Function"
UNIFORM_DIST_FUNCTION = "Uniform Distribution Function"
EXPONENTIAL_DIST_FUNCTION = "Exponential Distribution Function"
GAMMA_DIST_FUNCTION = "Gamma Distribution Function"
WALD_DIST_FUNCTION = "Wald Distribution Function"
DRIFT_DIFFUSION_ANALYTICAL_FUNCTION = "Drift Diffusion Analytical Function"

# ObjectiveFunctions
STABILITY_FUNCTION = 'Stability Function'
DISTANCE_FUNCTION = 'Distance Function'

# Interface Functions:
PORT_MAP_FUNCTION = 'Port Map Function'

#endregion

#region ---------------------------------------    SYSTEM / COMPOSITION   ----------------------------------------------

SYSTEM = "System"
SCHEDULER = "scheduler"
SYSTEM_INIT = 'System.__init__'
DEFAULT_SYSTEM = "DefaultSystem"
CONTROLLER = "controller"
ENABLE_CONTROLLER = "enable_controller"

RUN = 'run'
TRIAL = 'trial'

ROLES = 'roles'
CONDITIONS = 'conditions'
VALUES = 'values'
FUNCTIONS = 'functions'

#endregion

#region ---------------------------------------    AUTODIFF COMPOSITION   ----------------------------------------------

TRAINING_SET = 'training set'

#endregion

#region ----------------------------------------------    PROCESS   ----------------------------------------------------

PROCESS = "PROCESS"
PROCESSES = "processes"
PROCESS_INIT = 'Process.__init__'
PATHWAY = "pathway"
CLAMP_INPUT = "clamp_input"
SOFT_CLAMP = "soft_clamp"
HARD_CLAMP = "hard_clamp"
PULSE_CLAMP = "pulse_clamp"
NO_CLAMP = "no_clamp"
LEARNING_RATE = "learning_rate"
CONTROL = 'CONTROL'
GATING = 'gating'
PROCESS_DEFAULT_PROJECTION_FUNCTION = "Default Projection Function"
PROCESS_EXECUTE = "ProcessExecute"
MECHANISM_EXECUTED_LOG_ENTRY = "Mechanism Executed"
#endregion

#region ---------------------------------------------    MECHANISM   ---------------------------------------------------

MECHANISM = 'MECHANISM'
MECHANISMS = 'MECHANISMS'
MECHANISM_NAME = "MECHANISM NAME"
MECHANISM_DEFAULT = "DEFAULT MECHANISM"
DEFAULT_PROCESSING_MECHANISM = "DefaultProcessingMechanism"
PROCESS_DEFAULT_MECHANISM = "ProcessDefaultMechanism"
MECHANISM_TYPE = "Mechanism Type" # Used in mechanism dict specification (e.g., in process.pathway[])
MECHANISM_DEFAULTInputValue = "Mechanism Default Input Value " # Used in mechanism specification dict
MECHANISM_PARAM_VALUE = "Mechanism Parameter Value"                 # Used to specify mechanism param value
MECHANISM_DEFAULTParams = "Mechanism Default Parameters"           # Used in mechanism specification dict
CONDITION = 'condition'

# Keywords for OUTPUT_PORT_VARIABLE dict:
OWNER_VARIABLE = 'OWNER_VARIABLE'
OWNER_VALUE = 'OWNER_VALUE'
OWNER_EXECUTION_COUNT = 'EXECUTION_COUNT'
OWNER_EXECUTION_TIME = 'EXECUTION_TIME'
INPUT_PORT_VARIABLES = 'INPUT_PORT_VARIABLES'
PARAMS_DICT = 'PARAMS_DICT'

# this exists because keywords like OWNER_VALUE are set as properties
# on Mechanisms, so you can't just change the string value to map
# as you want - the property "value" will be overwritten then
output_port_spec_to_parameter_name = {
    OWNER_VARIABLE: VARIABLE,
    OWNER_VALUE: VALUE,
    INPUT_PORT_VARIABLES: 'input_port_variables',
    OWNER_EXECUTION_COUNT: 'execution_count',
    OWNER_EXECUTION_TIME: 'current_execution_time'
}

# Dictionaries of labels for input, output and target arrays
INPUT_LABELS_DICT = 'input_labels_dict'
OUTPUT_LABELS_DICT = 'output_labels_dict'
TARGET_LABELS_DICT = 'target_labels_dict'

ORIGIN = 'ORIGIN'
INTERNAL = 'INTERNAL'
CYCLE = 'CYCLE'
INITIALIZE_CYCLE = 'INITIALIZE_CYCLE'
# AUTO_LEARNING = 'AUTO_LEARNING'
TERMINAL = 'TERMINAL'
SINGLETON = 'ORIGIN AND TERMINAL'
LEARNING = 'LEARNING'
SAMPLE = 'SAMPLE'
TARGET = 'TARGET'
ERROR = 'ERROR'

RESULTS = 'RESULTS'
RESULT = 'RESULT'
OUTPUT_MEAN = 'OUTPUT_MEAN'
OUTPUT_MEDIAN = 'OUTPUT_MEDIAN'
OUTPUT_VARIANCE = 'OUTPUT_VARIANCE'
OUTPUT_STD_DEV = 'OUTPUT_STD_DEV'
MECHANISM_VALUE = 'MECHANISM_VALUE'
SIZE = 'size'
K_VALUE = 'k_value'
THRESHOLD = 'threshold'
RATIO = 'ratio'

port_value = "Port value"   # Used in Port specification dict                                                #  to specify Port value
Port_ParamS = "Port params" # Used in Port specification dict

#endregion

#region ----------------------------------------    MODULATORY MECHANISMS ----------------------------------------------

# LearningMechanism:
LEARNING_SIGNALS = 'learning_signals'
LEARNING_SIGNAL_SPECS = 'LEARNING_SIGNAL_SPECS'
LEARNED_PARAM = 'learned_param'
LEARNED_PROJECTION = 'learned_projection'

# ControlMechanism / EVCControlMechanism / ObjectiveMechanism
SIMULATIONS = 'simulations'
OBJECTIVE_MECHANISM = "objective_mechanism"
OUTCOME = 'OUTCOME'
MONITOR = "monitor"
MONITOR_FOR_CONTROL = "monitor_for_control"
PREDICTION_MECHANISM = "Prediction Mechanism"
PREDICTION_MECHANISMS = "prediction_mechanisms"
PREDICTION_MECHANISM_TYPE = "prediction_mechanism_type"
PREDICTION_MECHANISM_PARAMS = "prediction_mechanism_params"
PREDICTION_MECHANISM_OUTPUT = "PredictionMechanismOutput"

MODULATORY_SIGNALS = 'modulatory_signals'
CONTROL_SIGNALS = 'control_signals'
CONTROL_SIGNAL_SPECS = 'CONTROL_SIGNAL_SPECS'
CONTROLLED_PARAMS = 'CONTROLLED_PARAMS'
CONTROL_PROJECTIONS = 'ControlProjections'
GATING_SIGNALS = 'gating_signals'
OUTCOME_FUNCTION = 'outcome_function'
COST_FUNCTION = 'cost_function'
COMBINE_OUTCOME_AND_COST_FUNCTION = 'combine_outcome_and_cost_function'
VALUE_FUNCTION = 'value_function'
SAVE_ALL_VALUES_AND_POLICIES = 'save_all_values_and_policies'
SYSTEM_DEFAULT_CONTROLLER = "DefaultController"
EVC_SIMULATION = 'CONTROL SIMULATION'
ALLOCATION_SAMPLES = "allocation_samples"


# GatingMechanism
MAKE_DEFAULT_GATING_MECHANISM = "make_default_gating_mechanism"
GATING_SIGNALS = 'gating_signals'
GATING_SIGNAL_SPECS = 'GATING_SIGNAL_SPECS'
GATE = 'GATE'
GATED_PORTS = 'GATED_PORTS'
GATING_PROJECTIONS = 'GatingProjections'
GATING_ALLOCATION = 'gating_allocation'

MODULATORY_SPEC_KEYWORDS = {LEARNING, LEARNING_SIGNAL, LEARNING_PROJECTION, LEARNING_MECHANISM,
                            CONTROL, CONTROL_SIGNAL, CONTROL_PROJECTION, CONTROL_MECHANISM,
                            GATING, GATING_SIGNAL, GATING_PROJECTION, GATING_MECHANISM}

MODULATED_PARAMETER_PREFIX = 'mod_'

#endregion

#region ----------------------------------------------    PORTS  ------------------------------------------------------

PORT = "Port"
PORT_TYPE = "port_type"
# These are used as keys in Port specification dictionaries
PORTS = "PORTS"
MODULATES = "modulates"
PROJECTIONS = "projections"  # Used to specify projection list to Port DEPRECATED;  REPLACED BY CONNECTIONS
CONNECTIONS = 'CONNECTIONS'
Port_Name = "PortName"
PORT_PREFS = "PortPrefs"
PORT_CONTEXT = "PortContext"
ADD_INPUT_PORT = 'AddNewInputPort'     # Used by Mechanism._add_projection_to()
ADD_OUTPUT_PORT = 'AddNewOutputPort'   # Used by Mechanism._add_projection_from()
FULL = 'FULL'
OWNER = 'owner'
REFERENCE_VALUE = 'reference_value'
REFERENCE_VALUE_NAME = 'reference_value_name'

# InputPorts:
PRIMARY = 'Primary'
INPUT_PORTS = 'input_ports'
INPUT_PORT_PARAMS = 'input_port_params'
WEIGHT = 'weight'
EXPONENT = 'exponent'
INTERNAL_ONLY = 'internal_only'

# ParameterPorts:
PARAMETER_PORTS = 'parameter_ports'
PARAMETER_PORT_PARAMS = 'parameter_port_params'

# OutputPorts:
OUTPUT_PORTS = 'output_ports'
OUTPUT_PORT_PARAMS = 'output_ports_params'
STANDARD_OUTPUT_PORTS = 'standard_output_ports'
INDEX = 'index'       # For backward compatibility with INDEX and ASSIGN
ASSIGN = 'assign'     # For backward compatibility with INDEX and ASSIGN
CALCULATE = 'assign'  # For backward compatibility with CALCULATE

# Modulation
MODULATION = 'modulation'
MONITOR_FOR_MODULATION = 'monitor_for_modulation'
ADDITIVE = ADDITIVE_PARAM = 'additive_param'
MULTIPLICATIVE = MULTIPLICATIVE_PARAM = 'multiplicative_param'
OVERRIDE = OVERRIDE_PARAM = 'OVERRIDE'
DISABLE = DISABLE_PARAM = 'DISABLE'

#endregion

#region ---------------------------------------------    PROJECTION  ---------------------------------------------------

# Attributes / KVO keypaths / Parameters
PROJECTION = "Projection"
PROJECTION_TYPE = "PROJECTION_TYPE"
PROJECTION_PARAMS = "ProjectionParams"
MAPPING_PROJECTION_PARAMS = "MappingProjectionParams"
LEARNING_PROJECTION_PARAMS = 'LearningProjectionParams'
CONTROL_PROJECTION_PARAMS = "ControlProjectionParams"
GATING_PROJECTION_PARAMS = 'GatingProjectionParams'
PROJECTION_SENDER = 'projection_sender'
SENDER = 'sender'
RECEIVER = "receiver"
PROJECTION_DIRECTION = {SENDER: 'to',
                        RECEIVER: 'from'}
RECEIVER_ARG = 'receiver'
MONITOR_FOR_LEARNING = 'monitor_for_learning'
AUTO = 'auto'
HETERO = 'hetero'


#endregion

#region ----------------------------------------------    FUNCTION   ---------------------------------------------------


# General ------------------------------------------------

FUNCTION_PARAMETER_PREFIX = 'func_'
CUSTOM_FUNCTION = 'custom_function'
FUNCTION_OUTPUT_TYPE = 'output_type'
OUTPUT_TYPE = 'output'
OVERWRITE = 'overwrite'
REINITIALIZE = "reinitialize"
REINITIALIZE_WHEN = "reinitialize_when"

LOW = 'low'
HIGH = 'high'
BOUNDS = 'bounds'
MODE = 'mode'
REST = "rest"

# Function-specific ---------------------------------------

STATEFUL_FUNCTION = 'stateful_function'
INTEGRATOR_FUNCTION = 'integrator_function'
MEMORY_FUNCTION = 'memory_function'
HAS_INITIALIZERS='has_initializers'
INCREMENT = 'increment'
INTEGRATION_TYPE = "integration_type"
TIME_STEP_SIZE = 'time_step_size'
DECAY = 'decay'
INTEGRATOR_MODE_VALUE = "integrator_mode_value"
INSTANTANEOUS_MODE_VALUE = "instantaneous_mode_value"
LINEAR = 'linear'
CONSTANT = 'constant'
SIMPLE = 'scaled'
ADAPTIVE = 'modulatory'
DIFFUSION = 'diffusion'
EXPONENTIAL = 'exponential'
GAUSSIAN = 'gaussian'
SINUSOID = 'sinusoid'

COMBINE = 'combine'
SUM = 'sum'
DIFFERENCE = DIFFERENCE # Defined above for DISTANCE_METRICS
PRODUCT = 'product'
QUOTIENT = 'quotient'
SUBTRACTION = 'subtraction'
DIVISION = 'division'
SCALAR = 'scalar'
VECTOR = 'vector'
ARRANGEMENT = 'arrangement'

GAIN = 'gain'
BIAS = 'bias'
X_0 = "x_0"
LEAK = 'leak'
SLOPE = 'slope'
INTERCEPT = 'intercept'
RATE = 'rate'
SCALE = 'scale'
NOISE = 'noise'

BETA = 'beta'

DIST_SHAPE = 'dist_shape'

STANDARD_DEVIATION = 'standard_deviation'
VARIANCE = 'variance'
DIST_MEAN = 'mean'

MAX_VAL = 'MAX_VAL'
MAX_ABS_VAL = 'MAX_ABS_VAL'
MAX_INDICATOR = 'MAX_INDICATOR'
MAX_ABS_INDICATOR = 'MAX_ABS_INDICATOR'
MIN_VAL = 'MIN_VAL'
MIN_ABS_VAL = 'MIN_ABS_VAL'
MIN_INDICATOR = 'MIN_INDICATOR'
MIN_ABS_INDICATOR = 'MIN_ABS_INDICATOR'
PROB = 'PROB'
PROB_INDICATOR = 'PROB_INDICATOR'
MUTUAL_ENTROPY = 'mutual entropy'
PER_ITEM = 'per_item'

INITIALIZER = 'initializer'
INITIAL_V = 'initial_v'
INITIAL_W = 'initial_w'
WEIGHTS = "weights"
EXPONENTS = "exponents"
OPERATION = "operation"
OFFSET = "offset"

REWARD = 'reward'
NETWORK = 'network'

GAMMA = 'gamma'

MSE = 'MSE'
SSE = 'SSE'
#endregion

# model spec keywords
MODEL_SPEC_ID_TYPE = 'type'
MODEL_SPEC_ID_PSYNEULINK = 'PNL'
MODEL_SPEC_ID_GENERIC = 'generic'

MODEL_SPEC_ID_INPUT_PORTS = 'input_ports'
MODEL_SPEC_ID_OUTPUT_PORTS = 'output_ports'

MODEL_SPEC_ID_SENDER_MECH = 'sender'
MODEL_SPEC_ID_SENDER_PORT = 'sender_port'
MODEL_SPEC_ID_RECEIVER_MECH = 'receiver'
MODEL_SPEC_ID_RECEIVER_PORT = 'receiver_port'

MODEL_SPEC_ID_PARAMETER_SOURCE = 'source'
MODEL_SPEC_ID_PARAMETER_VALUE = 'value'

MODEL_SPEC_ID_NODES = 'nodes'
MODEL_SPEC_ID_PROJECTIONS = 'edges'
MODEL_SPEC_ID_COMPOSITION = 'graphs'



import matplotlib.pyplot as plt
import numpy as np
import psyneulink as pnl

# This implements the model by Cohen, J. D., & Huston, T. A. (1994). Progress in the use of interactive
# models for understanding attention and performance. In C. Umilta & M. Moscovitch(Eds.),
# AttentionandperformanceXV(pp.453-456). Cam- bridge, MA: MIT Press.
# The model aims to capute top-down effects of selective attention and the bottom-up effects of attentional capture.

# Define Variables ----------------------------------------------------------------------------------------------------
import psyneulink.core.components.functions.transferfunctions

rate = 0.1          # modified from the original code from 0.01 to 0.1
inhibition = -2.0   # lateral inhibition
bias = 4.0          # bias is positive since Logistic equation has - sing already implemented
threshold = 0.55    # modified from thr original code from 0.6 to 0.55 because incongruent condition won't reach 0.6
settle_trials = 50  # cycles until model settles

# Create mechanisms ---------------------------------------------------------------------------------------------------
#   Linear input units, colors: ('red', 'green'), words: ('RED','GREEN')
colors_input_layer = pnl.TransferMechanism(
    size=3,
    function=psyneulink.core.components.functions.transferfunctions.Linear,
    name='COLORS_INPUT'
)

words_input_layer = pnl.TransferMechanism(
    size=3,
    function=psyneulink.core.components.functions.transferfunctions.Linear,
    name='WORDS_INPUT'
)

task_input_layer = pnl.TransferMechanism(
    size=2,
    function=psyneulink.core.components.functions.transferfunctions.Linear,
    name='TASK_INPUT'
)

#   Task layer, tasks: ('name the color', 'read the word')
task_layer = pnl.RecurrentTransferMechanism(
    size=2,
    function=psyneulink.core.components.functions.transferfunctions.Logistic(),
    hetero=-2,
    integrator_mode=True,
    integration_rate=0.1,
    name='TASK'
)

#   Hidden layer units, colors: ('red','green') words: ('RED','GREEN')
colors_hidden_layer = pnl.RecurrentTransferMechanism(
    size=3,
    function=psyneulink.core.components.functions.transferfunctions.Logistic(x_0=4.0),
    integrator_mode=True,
    hetero=-2.0,
    # noise=pnl.NormalDist(mean=0.0, standard_deviation=.0).function,
    integration_rate=0.1,  # cohen-huston text says 0.01
    name='COLORS HIDDEN'
)

words_hidden_layer = pnl.RecurrentTransferMechanism(
    size=3,
    function=psyneulink.core.components.functions.transferfunctions.Logistic(x_0=4.0),
    hetero=-2,
    integrator_mode=True,
    # noise=pnl.NormalDist(mean=0.0, standard_deviation=.05).function,
    integration_rate=0.1,
    name='WORDS HIDDEN'
)
#   Response layer, responses: ('red', 'green'): RecurrentTransferMechanism for self inhibition matrix
response_layer = pnl.RecurrentTransferMechanism(
    size=2,
    function=psyneulink.core.components.functions.transferfunctions.Logistic(),
    hetero=-2.0,
    integrator_mode=True,
    integration_rate=0.1,
    name='RESPONSE'
)

# Log mechanisms ------------------------------------------------------------------------------------------------------
task_layer.set_log_conditions('value')
colors_hidden_layer.set_log_conditions('value')
words_hidden_layer.set_log_conditions('value')
response_layer.set_log_conditions('value')
# Connect mechanisms --------------------------------------------------------------------------------------------------
# (note that response layer projections are set to all zero first for initialization

color_input_weights = pnl.MappingProjection(
    matrix=np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0]
    ])
)

word_input_weights = pnl.MappingProjection(
    matrix=np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0]
    ])
)

task_input_weights = pnl.MappingProjection(
    matrix=np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ])
)

color_task_weights = pnl.MappingProjection(
    matrix=np.array([
        [4.0, 0.0],
        [4.0, 0.0],
        [4.0, 0.0]
    ])
)

task_color_weights = pnl.MappingProjection(
    matrix=np.array([
        [4.0, 4.0, 4.0],
        [0.0, 0.0, 0.0]
    ])
)

word_task_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 4.0],
        [0.0, 4.0],
        [0.0, 4.0]
    ])
)

task_word_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 0.0, 0.0],
        [4.0, 4.0, 4.0]
    ])
)

response_color_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0]
    ])
)

response_word_weights = pnl.MappingProjection(
    matrix=np.array([
        [0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0]
    ])
)

color_response_weights = pnl.MappingProjection(
    matrix=np.array([
        [1.5, 0.0],
        [0.0, 1.5],
        [0.0, 0.0]
    ])
)
word_response_weights = pnl.MappingProjection(
    matrix=np.array([
        [2.5, 0.0],
        [0.0, 2.5],
        [0.0, 0.0]
    ])
)
#
# Create pathways -----------------------------------------------------------------------------------------------------
color_response_process_1 = pnl.Process(
    pathway=[
        colors_input_layer,
        color_input_weights,
        colors_hidden_layer,
        color_response_weights,
        response_layer,
    ],
    name='COLORS_RESPONSE_PROCESS_1'
)

color_response_process_2 = pnl.Process(
    pathway=[
        response_layer,
        response_color_weights,
        colors_hidden_layer
    ],
    name='COLORS_RESPONSE_PROCESS_2'
)

word_response_process_1 = pnl.Process(
    pathway=[
        words_input_layer,
        word_input_weights,
        words_hidden_layer,
        word_response_weights,
        response_layer
    ],
    name='WORDS_RESPONSE_PROCESS_1'
)

word_response_process_2 = pnl.Process(
    pathway=[
        response_layer,
        response_word_weights,
        words_hidden_layer
    ],
    name='WORDS_RESPONSE_PROCESS_2'
)

task_color_response_process_1 = pnl.Process(
    pathway=[
        task_input_layer,
        task_input_weights,
        task_layer,
        task_color_weights,
        colors_hidden_layer])

task_color_response_process_2 = pnl.Process(
    pathway=[
        colors_hidden_layer,
        color_task_weights,
        task_layer])

task_word_response_process_1 = pnl.Process(
    pathway=[
        task_input_layer,
        task_layer,
        task_word_weights,
        words_hidden_layer])

task_word_response_process_2 = pnl.Process(
    pathway=[
        words_hidden_layer,
        word_task_weights,
        task_layer])


# Create system -------------------------------------------------------------------------------------------------------
Bidirectional_Stroop = pnl.System(
    processes=[
        color_response_process_1,
        word_response_process_1,
        task_color_response_process_1,
        task_word_response_process_1,
        color_response_process_2,
        word_response_process_2,
        task_color_response_process_2,
        task_word_response_process_2
    ],
    reinitialize_mechanisms_when=pnl.Never(),
    name='Bidirectional_Stroop_SYSTEM'
)
response_layer._add_system(Bidirectional_Stroop, pnl.TERMINAL)
Bidirectional_Stroop.terminal_mechanisms.append(response_layer)

input_dict = {colors_input_layer: [0, 0, 0],
              words_input_layer: [0, 0, 0],
              task_input_layer: [0, 1]}
print("\n\n\n\n")
print(Bidirectional_Stroop.run(inputs=input_dict))

for node in Bidirectional_Stroop.mechanisms:
    print(node.name, " Value: ", node.get_output_values(Bidirectional_Stroop))
# # LOGGING:
# colors_hidden_layer.set_log_conditions('value')
# words_hidden_layer.set_log_conditions('value')
#
# Bidirectional_Stroop.show()
# Bidirectional_Stroop.show_graph(show_dimensions=pnl.ALL)  # ,show_mechanism_structure=pnl.VALUES) # Uncomment to show graph of the system
#
# # Create threshold function -------------------------------------------------------------------------------------------
#
#
# def pass_threshold(response_layer, thresh):
#     results1 = response_layer.output_ports.values[0][0]  # red response
#     results2 = response_layer.output_ports.values[0][1]  # green response
#     if results1 >= thresh or results2 >= thresh:
#         return True
#     return False
#
#
# terminate_trial = {
#     pnl.TimeScale.TRIAL: pnl.While(pass_threshold, response_layer, threshold)
# }
#
# # Create test trials function -----------------------------------------------------------------------------------------
# # a BLUE word input is [1,0] to words_input_layer and GREEN word is [0,1]
# # a blue color input is [1,0] to colors_input_layer and green color is [0,1]
# # a color-naming trial is [1,0] to task_layer and a word-reading trial is [0,1]
#
#
# def trial_dict(red_color, green_color, neutral_color, red_word, green_word, neutral_word, CN, WR):
#
#     trialdict = {
#         colors_input_layer: [red_color, green_color, neutral_color],
#         words_input_layer: [red_word, green_word, neutral_word],
#         task_input_layer: [CN, WR]
#     }
#     return trialdict
#
#
# # Define initialization trials separately
# WR_initialize_input = trial_dict(0, 0, 0, 0, 0, 0, 0, 1)
# CN_initialize_input = trial_dict(0, 0, 0, 0, 0, 0, 1, 0)
#
# CN_incongruent_trial_input = trial_dict(1, 0, 0, 0, 1, 0, 1, 0)  # red_color, green color, red_word, green word, CN, WR
# CN_congruent_trial_input = trial_dict(1, 0, 0, 1, 0, 0, 1, 0)  # red_color, green color, red_word, green word, CN, WR
# CN_control_trial_input = trial_dict(1, 0, 0, 0, 0, 0, 1, 0)  # red_color, green color, red_word, green word, CN, WR
#
# WR_congruent_trial_input = trial_dict(1, 0, 0, 1, 0, 0, 0, 1)  # red_color, green color, red_word, green word, CN, WR
# WR_incongruent_trial_input = trial_dict(1, 0, 0, 0, 1, 0, 0, 1)  # red_color, green color, red_word, green word, CN, WR
# WR_control_trial_input = trial_dict(0, 0, 0, 1, 0, 0, 0, 1)  # red_color, green color, red_word, green word, CN, WR
#
# Stimulus = [[CN_initialize_input, CN_control_trial_input],
#             [CN_initialize_input, CN_incongruent_trial_input],
#             [CN_initialize_input, CN_congruent_trial_input]]
#
# Stimulus2 = [[WR_initialize_input, WR_control_trial_input],
#              [WR_initialize_input, WR_incongruent_trial_input],
#              [WR_initialize_input, WR_control_trial_input]]
#
# conditions = 3
# response_all = []
# response_all2 = []
#
# # Run color naming trials ----------------------------------------------------------------------------------------------
# for cond in range(conditions):
#     response_color_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [0.0, 0.0, 0.0],
#             [0.0, 0.0, 0.0]
#         ])
#     )
#     response_word_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [0.0, 0.0, 0.0],
#             [0.0, 0.0, 0.0]
#         ])
#     )
#
#     Bidirectional_Stroop.run(inputs=Stimulus[cond][0], num_trials=settle_trials)
#     response_color_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [1.5, 0.0, 0.0],
#             [0.0, 1.5, 0.0]
#         ])
#     )
#     response_word_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [2.5, 0.0, 0.0],
#             [0.0, 2.5, 0.0]
#         ])
#     )
#     Bidirectional_Stroop.run(inputs=Stimulus[cond][1], termination_processing=terminate_trial)
#
#     # Store values from run -----------------------------------------------------------------------------------------------
#     r = response_layer.log.nparray_dictionary('value')       # Log response output from special logistic function
#     rr = r['value']
#     n_r = rr.shape[0]
#     rrr = rr.reshape(n_r, 2)
#     response_all.append(rrr)  # .shape[0])
#     response_all2.append(rrr.shape[0])
#
#     # Clear log & reinitialize ----------------------------------------------------------------------------------------
#     response_layer.log.clear_entries(delete_entry=False)
#     colors_hidden_layer.log.clear_entries(delete_entry=False)
#     words_hidden_layer.log.clear_entries(delete_entry=False)
#     task_layer.log.clear_entries(delete_entry=False)
#     colors_hidden_layer.reinitialize([[0, 0, 0]])
#     words_hidden_layer.reinitialize([[0, 0, 0]])
#     response_layer.reinitialize([[0, 0]])
#     task_layer.reinitialize([[0, 0]])
#     print('response_all: ', response_all)
#
# # Run color naming trials ----------------------------------------------------------------------------------------------
# response_all3 = []
# response_all4 = []
# for cond in range(conditions):
#     response_color_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [0.0, 0.0, 0.0],
#             [0.0, 0.0, 0.0]
#         ])
#     )
#     response_word_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [0.0, 0.0, 0.0],
#             [0.0, 0.0, 0.0]
#         ])
#     )
#
#     Bidirectional_Stroop.run(inputs=Stimulus2[cond][0], num_trials=settle_trials)
#     response_color_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [1.5, 0.0, 0.0],
#             [0.0, 1.5, 0.0]
#         ])
#     )
#     response_word_weights = pnl.MappingProjection(
#         matrix=np.array([
#             [2.5, 0.0, 0.0],
#             [0.0, 2.5, 0.0]
#         ])
#     )
#     Bidirectional_Stroop.run(inputs=Stimulus2[cond][1], termination_processing=terminate_trial)
#
#     # Store values from run -----------------------------------------------------------------------------------------------
#     r2 = response_layer.log.nparray_dictionary('value')       # Log response output from special logistic function
#     rr2 = r2['value']
#     n_r2 = rr2.shape[0]
#     rrr2 = rr2.reshape(n_r2, 2)
#     response_all3.append(rrr2)  # .shape[0])
#     response_all4.append(rrr2.shape[0])
#
#     # Clear log & reinitialize ----------------------------------------------------------------------------------------
#     response_layer.log.clear_entries(delete_entry=False)
#     colors_hidden_layer.log.clear_entries(delete_entry=False)
#     words_hidden_layer.log.clear_entries(delete_entry=False)
#     task_layer.log.clear_entries(delete_entry=False)
#     colors_hidden_layer.reinitialize([[0, 0, 0]])
#     words_hidden_layer.reinitialize([[0, 0, 0]])
#     response_layer.reinitialize([[0, 0]])
#     task_layer.reinitialize([[0, 0]])
#     print('response_all: ', response_all)
#
#
# # Plot results --------------------------------------------------------------------------------------------------------
# # First, plot response layer activity for whole run
# plt.figure()
# # color naming plot
# plt.plot(response_all[0])
# plt.plot(response_all[1])
# plt.plot(response_all[2])
# # word reading plot
# plt.plot(response_all3[0])
# plt.plot(response_all3[1])
# plt.plot(response_all3[2])
# plt.show()
# # Second, plot regression plot
# # regression
# reg = np.dot(response_all2, 5) + 115
# reg2 = np.dot(response_all4, 5) + 115
# plt.figure()
#
# plt.plot(reg, '-s')  # plot color naming
# plt.plot(reg2, '-or')  # plot word reading
# plt.title('GRAIN MODEL with bidirectional weights')
# plt.legend(['color naming', 'word reading'])
# plt.xticks(np.arange(3), ('control', 'incongruent', 'congruent'))
# plt.ylabel('reaction time in ms')
# plt.show()



import numpy as np
import psyneulink as pnl
import multitasking_nn_model

# Setting up defaults
DEFAULT_TRAIN_ITERATIONS = 500
DEFAULT_NUM_REPLICATIONS = 100

DEFAULT_NUM_DIMENSIONS = 3
DEFAULT_NUM_FEATURES_PER_DIMENSION = 4


def load_or_create_patterns(num_dimensions=DEFAULT_NUM_DIMENSIONS,
                            num_features=DEFAULT_NUM_FEATURES_PER_DIMENSION):
    # TODO: get this code from Sebastian
    pass


def main():
    model = multitasking_nn_model.MultitaskingModel(DEFAULT_NUM_DIMENSIONS, DEFAULT_NUM_FEATURES_PER_DIMENSION)
    model.system.show_graph(show_dimensions=pnl.ALL, show_projection_labels=pnl.ALL) #show_processes=pnl.ALL)


if __name__ == '__main__':
    main()



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *******************************************  LearningAuxiliary ******************************************************

"""
.. _LearningAuxiliary_Overview:

Overview
--------

This module contains functions that automatically construct and compose the components necessary for learning
when this is specified for a process.

COMMENT:
    IMPLEMENT: LearningMechanism:
        PROCESS & SYSTEM:
          • Convert ProcessInputPort and SystemInputPort into Mechanisms with LinearFunction IDENTITY_FUNCTION
          • Use only one ObjectiveMechanism for all levels with the following args:
                default_variable[[ACTIVITY][ERROR]]
                monitored_output_ports: [[error_mech.OutputPort][error_mech.objective_mechanism.OutputPort]]
                names: [[ACTIVITY][ERROR]]
                function:  ErrorDerivative(variable, derivative)
                               variable[0] = activity
                               variable[1] = error_signal from error_mech ObjectiveMechanism (target for TERMINAL)
                               derivative = error_derivative (1 for TERMINAL)
                role:LEARNING
          • Use only one LearningMechanism with the following args:
                variable[[ACTIVATION_INPUT_INDEX][ACTIVATION_OUTPUT_INDEX][ERROR_SIGNAL_INDEX]
                activation_derivative
                error_matrix
                function
            Initialize and assign args with the following WIZZARD:
        WIZZARD:
            Needs to know
                activation_mech_output (Mechanism)
                    activation_derivative (function)
                error_mech (Mechanism)
                    error_derivative (function)
                    error_matrix (ndarray) - for MappingProjection from activation_mech_output to error_mech
            ObjectiveMechanism:
                Initialize variable:
                      use error_mech.outputPort.valuee to initialize variable[ACTIVITY]
                      use outputPort.value of error_mech's objective_mechanism to initialize variable[ERROR]
                Assign mapping projections:
                      nextLevel.outputPort.value -> input_ports[ACTIVITY] of ObjectiveMechanism
                      nextLevel.objective_mechanism.outputPort.value  -> input_ports[ERROR] of ObjectiveMechanism
                NOTE: For `TERMINAL` Mechanism:
                          error_mech is Process or System InputPort (function=Linear, so derivative =1), so that
                             error_mech.outputPort.value is the target, and
                             error_derivative = 1
                             error_matrix = IDENTITY_MATRIX (this should be imposed)
            LearningMechanism:
                Initialize variable:
                      use mapping_projection.sender.value to initialize variable[ACTIVATION_INPUT_INDEX]
                      use activation_mech_output_outputPort.value to initialize variable[ACTIVATION_OUTPUT_INDEX]
                      use error_mech.objecdtive_mechanism.OutputPort.value to initialize variable[ERROR_SIGNAL_INDEX]
                Assign activation_derivative using function of activation_mech_output of mapping_projection (one being learned)
                Assign error_derivative using function of error_mech
                Assign error_matrix as runtime_param using projection to error_mech [ALT: ADD TO VARIABLE]
                Assign mapping projections:
                      mapping_projection.sender -> input_ports[ACTIVATION_INPUT_INDEX] of LearningMechanism
                      activation_mech_output.outputPort -> input_ports[ACTIVATION_OUTPUT_INDEX] of LearningMechanism
                      error_mech.objective_mechanism.OutputPort.value -> input_ports[ERROR_SIGNAL_INDEX]

            For TARGET MECHANISM:  Matrix is IDENTITY MATRIX??
            For TARGET MECHANISM:  derivative for ObjectiveMechanism IDENTITY FUNCTION


    *************************************************************************

    Call in _instantiate_attributes_before_function() of LearningProjection

    Do the following:
        Get:
            activation_mech_projection (one being learned) (MappingProjection)
            activation_output_mech (ProcessingMechanism)
            activation_mech_input (OutputPort)
            activation_mech_output (OutputPort)
            activation_derivative (function)
            error_matrix (ParameterPort)
            error_derivative (function)
            error_mech (error_source_mech) (ProcessingMechanism)
            error_signal_mech (LearningMechanism or ObjectiveMechanism)
            error_signal_mech_output (OutputPort)
            error_objective_mech (error_source_objective_mech) (ObjectiveMechanism)

        Instantiate:
            ObjectiveMechanism:
                Construct with:
                   monitor_values: [error_mech.outputPort.value, error_mech.objective_mech.outputPort.value]
                   names = [ACTIVITY/SAMPLE, ERROR/TARGET]
                   function = ErrorDerivative(derivative=error_derivative)
                NOTE: For `TERMINAL` Mechanism:
                          error_mech is Process or System InputPort (function=Linear, so derivative =1), so that
                             error_mech.outputPort.value is the target, and
                             error_derivative = 1
                             error_matrix = IDENTITY_MATRIX (this should be imposed)
            LearningMechanism:
                Construct with:
                    variable=[[activation_mech_input],[activation_mech_output],[ObjectiveMechanism.outputPort.value]]
                    error_matrix = error_matrix
                    function = one specified with Learning specification
                               NOTE: can no longer be specified as function for LearningProjection
                    names = [ACTIVATION_INPUT, ACTIVATION_OUTPUT, ERROR_SIGNAL]
                        NOTE:  this needs to be implemented for LearningMechanism as it is for ObjectiveMechanism
                Check that, if learning function expects a derivative (in user_params), that the one specified
                    is compatible with the function of the activation_output_mech
                Assign:
                    NOTE:  should do these in their own Learning module function, called by LearningMechanaism directly
                        as is done for ObjectiveMechanism
                    MappingProjection: activation_mech_projection.sender -> LearningMechanism.input_port[ACTIVATION_INPUT]
                    MappingProjection: activation_output_mech.outputPort -> LearningMechanism.input_port[ACTIVATION_OUTPUT]
                    MappingProjection: ObjectiveMechanism -> LearningMechanism.input_port[ERROR_SIGNAL]


COMMENT


.. _LearningProjection_Class_Reference:

Class Reference
---------------

"""

import warnings

import numpy as np
import typecheck as tc

from psyneulink.core.components.component import function_type, method_type
from psyneulink.core.components.functions.combinationfunctions import PredictionErrorDeltaFunction
from psyneulink.core.components.functions.learningfunctions import BackPropagation, Hebbian, Reinforcement, TDLearning
from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.mechanisms.modulatory.learning.learningmechanism import ACTIVATION_INPUT, \
    ACTIVATION_OUTPUT, ERROR_SIGNAL, LearningMechanism, LearningType, LearningTiming
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.projections.modulatory.learningprojection import LearningProjection
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.components.shellclasses import Function
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.components.ports.parameterport import ParameterPort
from psyneulink.core.globals.context import Context, ContextFlags
from psyneulink.core.globals.keywords import \
    BACKPROPAGATION_FUNCTION, COMPARATOR_MECHANISM, HEBBIAN_FUNCTION, IDENTITY_MATRIX, LEARNING, LEARNING_MECHANISM, \
    MATRIX, MONITOR_FOR_LEARNING, NAME, OUTCOME, PREDICTION_ERROR_MECHANISM, PROJECTIONS, RL_FUNCTION, SAMPLE, \
    TARGET, TDLEARNING_FUNCTION, VARIABLE, WEIGHT
from psyneulink.library.components.mechanisms.processing.objective.predictionerrormechanism import PredictionErrorMechanism

__all__ = [
    'LearningAuxiliaryError'
]


class LearningAuxiliaryError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value


def _instantiate_learning_components(learning_projection, context=None):
    """Instantiate learning components for a LearningProjection

    Instantiates a LearningMechanism or ObjectiveMechanism as the sender for each learning_projection in a learning
        sequence.  A learning sequence is defined as a sequence of ProcessingMechanisms, each of which has a
        projection — that has been specified for learning — to the next Mechanism in the sequence.  This method
        instantiates the components required to support learning for those projections (most importantly,
        the LearningMechanism that provide them with the learning_signal required to modify the matrix of the
        projection, and the ObjectiveMechanism that calculates the error_signal used to generate the learning_signals).


    It instantiates a LearningMechanism or ObjectiveMechanism as the sender for the learning_projection:
    - a LearningMechanism for projections to any ProcessingMechanism that is not the last in the learning sequence;
    - an ObjectiveMechanism for projections to a ProcessingMechanism that is the last in the learning sequence

    Assume that learning_projection's variable and parameters have been specified and validated,
       (which is the case when this method is called from the learning_projection itself in _instantiate_sender()).

    Notes:

    * Once the `receiver` for the learning_projection has been identified, or instantiated:
        - it is thereafter referred to (by reference to it owner) as the `activation_mech_projection`,
            and the Mechanism to which it projects as the `activation_output_mech`;
        - the Mechanism to which it projects is referred referred to as the error_mech (the source of the error_signal).

    * See LearningComponents class for the names of the components of learning used here.

    * This method supports only a single pathway for learning;  that is, the learning sequence must be a linear
        sequence of ProcessingMechanisms.  This is consistent with its implementation at the Process level;
        convergent and divergent pathways for learning can be accomplished through Composition in a
        System.  Accordingly:

            - each LearningMechanism can have only one LearningProjection
            - each ProcessingMechanism can have only one MappingProjection that is subject to learning

      When searching downstream for projections that are being learned (to identify the error_mech Mechanism as a
      source for the LearningMechanism's error_signal), the method uses the first projection that it finds, beginning
      in `primary outputPort <OutputPort_Primary>` of the activation_mech_output, and continues similarly
      in the error_mech.  If any conflicting implementations of learning components are encountered,
      an exception is raised.

    """
    # IMPLEMENTATION NOTE: CURRENT IMPLEMENTATION ONLY SUPPORTS CALL FROM LearningProjection._instantiate_sender();
    #                      IMPLEMENTATION OF SUPPORT FOR EXTERNAL CALLS:
    #                      - SHOULD CHECK UP FRONT WHETHER SENDER OR RECEIVER IS SPECIFIED, AND BRANCH ACCORDINGLY
    #                            FAILURE TO SPECIFY BOTH SHOULD RAISE AN EXCEPTION (OR FURTHER DEFER INSTANTIATION?)
    #                      - WILL REQUIRE MORE EXTENSIVE CHECKING AND VALIDATION
    #                              (E.G., OF WHETHER ANY LearningMechanism IDENTIFIED HAVE A PROJECTION FROM AN
    #                               APPROPRIATE ObjectiveMechanism, etc.
    from psyneulink.library.components.mechanisms.processing.objective.comparatormechanism import \
        ComparatorMechanism, MSE

    # Call should generally be from LearningProjection._instantiate_sender,
    #    but may be used more generally in the future
    if context.source != ContextFlags.METHOD:
        raise LearningAuxiliaryError("PROGRAM ERROR".format("_instantiate_learning_components only supports "
                                                             "calls from a LearningProjection._instantiate_sender()"))


    # First determine whether receiver for LearningProjection has been instantiated and, if not, instantiate it:
    #    this is required to instantiate the LearningMechanism, as well as an ObjectiveMechanism if necessary.
    try:
        isinstance(learning_projection.receiver.owner, MappingProjection)
    except AttributeError:
        _instantiate_receiver_for_learning_projection(learning_projection, context=context)
        # IMPLEMENTATION NOTE: THIS SHOULD IDENTIFY/CONFIRM RECEIVER BASED ON PROJECTIONS TO THE OBJECTIVE MECHANISM
        #                      OF THE learning_projection's LearningMechanism;  IF THERE ARE NONE, THEN RAISE EXCEPTION

    # Next, validate that the receiver does not receive any other LearningProjections
    # IMPLEMENTATION NOTE:  this may be overly restrictive in the context of a system --
    #                       will need to be dealt with in Composition implementation (by examining its learning graph??)
    if any((isinstance(projection, LearningProjection) and not projection is learning_projection) for projection in
           learning_projection.receiver.mod_afferents):
        raise LearningAuxiliaryError("{} can't be assigned as LearningProjection to {} since that already has one.".
                                     format(learning_projection.name, learning_projection.receiver.owner.name))

    # Now that the receiver has been identified, use it to get the components (lc) needed for learning.
    # Note: error-related components may not yet be defined (if LearningProjection is for a TERMINAL mechanism)
    lc = LearningComponents(learning_projection=learning_projection)

    # Check if activation_output_mech already has a projection to an ObjectiveMechanism or a LearningMechanism
    #    (i.e., the source of an error_signal for learning_projection)
    # IMPLEMENTATION NOTE: this uses the first projection found (starting with the primary outputPort)

    # FIX: CHECK THAT THE PROJECTION IS TO ANOTHER MECHANISM IN THE CURRENT PROCESS;
    #      OTHERWISE, ALLOW OBJECTIVE_MECHANISM TO BE IMPLEMENTED


    # Check whether activation_output_mech belongs to more than one process and, if it does, it is the TERMINAL
    #    the one currently being instantiated, in which case it needs to be assigned an ObjectiveMechanism
    #
    # If the only MappingProjections from activation_output_mech are to mechanisms that are not in the same
    #    process as the mechanism that projects to it (i.e., lc.activation_mech_input.owner), then:
    #    - set is_target to True so that it will be assigned an ObjectiveMechanism
    # Note: this handles the case in which the current mech belongs to more than one process, and is the
    #    the TERMINAL of the one currently being instantiated.
    # IMPLEMENTATION NOTE:  could check whether current mech is a TERMINAL for the process currently being
    #                       instantiated, identified by the mechanism that projects to it.  However, the latter
    #                       may also belong to more than one process, so would have to sort that out.  Current
    #                       implementation doesn't have to worry about that.

    objective_mechanism = None

    # if activation_output_mech has outgoing projections
    if (lc.activation_mech_output.efferents and
            # if the ProcessingMechanisms to which activation_mech_output projects do not belong to any of the same
            # processes to which the mechanisms that project to activation_output_mech belong, then it should be
            # treated as a TERMINAL and is_target should be set to True
            not any(
                        isinstance(projection.receiver.owner, ProcessingMechanism_Base) and
                        any(        # processes of ProcessingMechanisms to which activation_output_mech projects
                                    process in projection.receiver.owner.processes
                                    # processes of mechanism that project to activation_output_mech
                                    for process in lc.activation_mech_input.owner.processes)
                        for projection in lc.activation_mech_output.efferents)):
        is_target = True

    else:

        for projection in lc.activation_mech_output.efferents:

            receiver_mech = projection.receiver.owner
            receiver_state = projection.receiver

            # Check if projection already projects to a LearningMechanism or an ObjectiveMechanism

            # activation_output_mech projects to a LearningMechanism
            if isinstance(receiver_mech, LearningMechanism):

                # IMPLEMENTATION NOTE:  THIS IS A SANITY CHECK;  IF THE learning_projection ALREADY HAS A SENDER
                #                       THAT IS A LearningMechanism, THIS FUNCTION SHOULD NOT HAVE BEEN CALLED
                # If receiver_mech is a LearningMechanism that is the sender for the learning_projection,
                #    raise an exception since this function should not have been called.
                if learning_projection.sender is receiver_mech:
                    raise LearningAuxiliaryError("PROGRAM ERROR: "
                                                  "{} already has a LearningMechanism as its sender ({})".
                                                 format(learning_projection.name, receiver_mech.name))

                # If receiver_mech for activation_output_mech is a LearningMechanism that receives projections to its:
                #     - ACTIVATION_INPUT InputPort from activation_mech_input.owner
                #     - ACTIVATION_OUTPUT InputPort from activation_output_mech
                #         (i.e., the mechanism before activation_output_mech in the learning sequence)
                #         then this should be the LearningMechanism for the learning_projection,
                #         so issue warning, assign it as the sender, and return
                if (receiver_state.name is ACTIVATION_OUTPUT and
                        any(projection.sender.owner is lc.activation_mech_input.owner
                            for projection in receiver_mech.input_ports[ACTIVATION_INPUT].path_afferents)):
                        warnings.warn("An existing LearningMechanism ({}) was found for and is being assigned to {}".
                                      format(receiver_mech.name, learning_projection.name))
                        learning_projection.sender = receiver_mech
                        return

            # activation_output_mech already projects to an ObjectiveMechanism used for learning
            #    (presumably instantiated for another process);
            #    note:  doesn't matter if it is not being used for learning (then its just another ProcessingMechanism)
            elif isinstance(receiver_mech, ObjectiveMechanism) and LEARNING in receiver_mech._role:

                # ObjectiveMechanism is for learning but projection is not to its SAMPLE inputPort
                if LEARNING in receiver_mech._role and not receiver_state.name is SAMPLE:
                    raise LearningAuxiliaryError("PROGRAM ERROR: {} projects to the {} rather than the {} "
                                                  "inputPort of an ObjectiveMechanism for learning {}".
                                                 format(lc.activation_output_mech.name,
                                                         receiver_state.name,
                                                         SAMPLE,
                                                         receiver_mech.name))

                # IMPLEMENTATION NOTE:  THIS IS A SANITY CHECK;  IF THE learning_projection ALREADY HAS A SENDER
                #                       THAT IS A LearningMechanism, THIS FUNCTION SHOULD NOT HAVE BEEN CALLED
                # If the ObjectiveMechanism projects to a LearningMechanism that is the sender for the
                #     learning_projection, raise exception as this function should not have been called
                elif (isinstance(learning_projection.sender, LearningMechanism) and
                          any(learning_projection.sender.owner is projection.receiver.owner
                              for projection in receiver_mech.output_port.efferents)):
                    raise LearningAuxiliaryError("PROGRAM ERROR:  {} already has an "
                                                  "ObjectiveMechanism ({}) and a "
                                                  "LearningMechanism ({}) assigned to it".
                                                 format(learning_projection.name,
                                                         receiver_mech.name,
                                                         learning_projection.sender.owner.name))

                else:
                    objective_mechanism = receiver_mech


        # FIX: CHECK THAT THE PROJECTION IS TO ANOTHER MECHANISM IN THE CURRENT PROCESS;
        #      OTHERWISE, ALLOW OBJECTIVE_MECHANISM TO BE IMPLEMENTED
        #      BUT HOW CAN IT KNOW THIS, SINCE UNTIL THE METHOD IS PART OF A COMPOSITION, IT DOESN'T KNOW WHAT
        #          COMPOSITION IS BEING CALLED.  -- LOOK AT LearningProjection_OLD TO SEE HOW IT WAS HANDLED THERE.
        #          OR IS A TARGET/OBJECTIVE MECHANISM ALWAYS ASSIGNED FOR A PROCESS, AND THEN DELETED IN SYSTEM GRAPH?
        #          IN THAT CASE, JUST IGNORE SECOND CONDITION BELOW (THAT IT HAS NO PROJECTIONS TO LEARNING_PROJECTIONS?

        # Next, determine whether an ObjectiveMechanism or LearningMechanism should be assigned as the sender
        # It SHOULD be an ObjectiveMechanism (i.e., TARGET) if either:
        #     - it has no outgoing projections or
        #     - it has no projections that receive a LearningProjection;
        #   in either case, lc.error_projection returns None.
        #   Note: this assumes that LearningProjections are being assigned from the end of the pathway to the beginning.
        is_target = not lc.error_projection

    # INSTANTIATE learning function

    # Note: have to wait to do it here, as Backpropagation needs error_matrix,
    #       which depends on projection to ObjectiveMechanism

    # IMPLEMENTATION NOTE:
    #      THESE SHOULD BE MOVE TO COMPOSITION WHEN IT IS IPMLEMENTED
    #      THESE SHOULD BE MOVED (ALONG WITH THE SPECIFICATION FOR LEARNING) TO A DEDICATED LEARNING SPEC
    #      FOR PROCESS AND SYSTEM, RATHER THAN USING A LearningProjection
    # Get function used for learning and the learning_rate from their specification in the LearningProjection
    # FIXME: learning_function is deprecated
    learning_function = learning_projection._learning_function
    learning_rate = learning_projection._learning_rate
    error_function = learning_projection._error_function

    # HEBBIAN LEARNING FUNCTION
    if learning_function.componentName is HEBBIAN_FUNCTION:

        activation = np.zeros_like(lc.activation_mech_input.value)
        learning_rate = learning_projection.learning_function.learning_rate
        activation_output = error_signal = None

        # FIX: GET AND PASS ANY PARAMS ASSIGNED IN LearningProjection.learning_function ARG:
        # FIX:     ACTIVATION FUNCTION AND/OR LEARNING RATE
        learning_function = Hebbian(default_variable=activation,
                                    # activation_function=lc.activation_mech_fct,
                                    learning_rate=learning_rate)

        objective_mechanism = lc.activation_output_mech

    # REINFORCEMENT LEARNING FUNCTION
    elif learning_function.componentName is RL_FUNCTION:

        activation_input = np.zeros_like(lc.activation_mech_input.value)
        activation_output = np.zeros_like(lc.activation_mech_output.value)

        # Force output activity and error arrays to be scalars
        error_signal = np.array([0])
        error_output = np.array([0])
        learning_rate = learning_projection.learning_function.learning_rate

        # FIX: GET AND PASS ANY PARAMS ASSIGNED IN LearningProjection.learning_function ARG:
        # FIX:     ACTIVATION FUNCTION AND/OR LEARNING RATE
        learning_function = Reinforcement(default_variable=[activation_input, activation_output, error_signal],
                                          # activation_function=lc.activation_mech_fct,
                                          learning_rate=learning_rate)

    elif learning_function.componentName is TDLEARNING_FUNCTION:
        activation_input = np.zeros_like(lc.activation_mech_input.value)
        activation_output = np.zeros_like(lc.activation_mech_output.value)

        error_output = np.zeros_like(lc.activation_mech_output.value)
        error_signal = np.zeros_like(lc.activation_mech_output.value)
        learning_rate = learning_projection.learning_function.learning_rate

        learning_function = TDLearning(default_variable=[activation_input,
                                                         activation_output,
                                                         error_signal],
                                       # activation_function=lc.activation_mech_fct,
                                       learning_rate=learning_rate)

    # BACKPROPAGATION LEARNING FUNCTION
    elif learning_function.componentName is BACKPROPAGATION_FUNCTION:

        # Get activation_output_mech values
        activation_input = np.zeros_like(lc.activation_mech_input.value)
        activation_output = np.zeros_like(lc.activation_mech_output.value)

        # Validate that the function for activation_output_mech has a derivative
        try:
            activation_derivative = lc.activation_mech_fct.derivative
        except AttributeError:
            raise LearningAuxiliaryError("Function for activation_output_mech of {} "
                                          "must have a derivative to be used "
                                          "with {}".format(learning_projection.name,
                                                           BackPropagation.componentName))

        # Get error_mech values
        if is_target:
            error_output = np.ones_like(lc.activation_mech_output.value)
            error_signal = np.zeros_like(lc.activation_mech_output.value)
            error_matrix = np.identity(len(error_signal))
            # IMPLEMENTATION NOTE: Assign error_derivative to derivative of ProcessingInputPort or SystemInputPort
            #                      function when these are fully implemented as mechanisms
            # activation_derivative = Linear().derivative
            error_derivative = Linear().derivative

        else:
            error_output = np.zeros_like(lc.error_mech_output.value)
            error_signal = np.zeros_like(lc.error_signal_mech.output_ports[ERROR_SIGNAL].value)
            error_matrix = lc.error_matrix
            try:
                error_derivative = lc.error_derivative
            except AttributeError:
                raise LearningAuxiliaryError("Function for error_mech of {} "
                                              "must have a derivative to be "
                                              "used with {}".format(learning_projection.name,
                                                                    BackPropagation.componentName))

        # FIX: GET AND PASS ANY PARAMS ASSIGNED IN LearningProjection.learning_function ARG:
        # FIX:     DERIVATIVE, LEARNING_RATE, ERROR_MATRIX
        learning_function = BackPropagation(default_variable=[activation_input,
                                                              activation_output,
                                                              # error_output,
                                                              error_signal],
                                            activation_derivative_fct=activation_derivative,
                                            learning_rate=learning_rate)

    else:
        raise LearningAuxiliaryError("PROGRAM ERROR: unrecognized learning "
                                      "function ({}) for {}".format(learning_function.componentName,
                                                                    learning_projection.name))


    # INSTANTIATE ObjectiveMechanism

    # If it is a TARGET, instantiate an ObjectiveMechanism and assign as learning_projection's sender
    if is_target:

        if objective_mechanism is None:
            # Instantiate ObjectiveMechanism
            # Notes:
            # * MappingProjections for ObjectiveMechanism's input_ports will be assigned in its own call to Composition
            # * Need to specify both default_variable and monitored_output_ports since they may not be the same
            #    sizes (e.g., for RL the monitored_output_port for the sample may be a vector, but its input_value must be scalar)
            # SAMPLE inputPort for ObjectiveMechanism should come from activation_mech_output
            # TARGET inputPort for ObjectiveMechanism should be specified by string (TARGET),
            #     so that it is left free to later be assigned a projection from ProcessInputPort and/or SystemInputPort
            # Assign derivative of Linear to lc.error_derivative (as default, until TARGET projection is assigned);
            #    this will induce a simple subtraction of target-sample (i.e., implement a comparator)
            sample_input = target_input = error_output
            # MODIFIED 10/10/17 OLD:
            # objective_mechanism = ComparatorMechanism(sample=lc.activation_mech_output,
            #                                           target=TARGET,
            #                                           # input_ports=[sample_input, target_input],
            #                                           # FOR TESTING: ALTERNATIVE specifications of input_ports arg:
            #                                           # input_ports=[(sample_input, FULL_CONNECTIVITY_MATRIX),
            #                                           #               target_input],
            #                                           # input_ports=[(sample_input, RANDOM_CONNECTIVITY_MATRIX),
            #                                           #               target_input],
            #                                           input_ports=[{NAME:SAMPLE,
            #                                                          VARIABLE:sample_input,
            #                                                          WEIGHT:-1
            #                                                          },
            #                                                         {NAME:TARGET,
            #                                                          VARIABLE:target_input,
            #                                                          # WEIGHT:1
            #                                                          }],
            #                                           name="{} {}".format(lc.activation_output_mech.name,
            #                                                               COMPARATOR_MECHANISM))
            # MODIFIED 10/10/17 NEW:
            if learning_function.componentName == TDLEARNING_FUNCTION:
                objective_mechanism = PredictionErrorMechanism(
                        sample={NAME: SAMPLE,
                                VARIABLE: sample_input,
                                PROJECTIONS: [lc.activation_mech_output]},
                        target={NAME: TARGET,
                                VARIABLE: target_input},
                        function=PredictionErrorDeltaFunction(gamma=1.0),
                        name="{} {}".format(lc.activation_output_mech.name,
                                            PREDICTION_ERROR_MECHANISM))
            else:
                objective_mechanism = ComparatorMechanism(sample={NAME: SAMPLE,
                                                                  VARIABLE: sample_input,
                                                                  PROJECTIONS: [lc.activation_mech_output],
                                                                  WEIGHT: -1},
                                                          target={NAME: TARGET,
                                                                  VARIABLE: target_input},
                                                          function=error_function,
                                                          output_ports=[OUTCOME, MSE],
                                                          name="{} {}".format(lc.activation_output_mech.name,
                                                                              COMPARATOR_MECHANISM),
                                                          context=context)
                # MODIFIED 10/10/17 END

            # # FOR TESTING: ALTERNATIVE to Direct call to ObjectiveMechanism
            # #              (should produce identical result to use of ComparatorMechanism above)
            # objective_mechanism = ObjectiveMechanism(monitored_output_ports=[lc.activation_mech_output,
            #                                                            TARGET],
            #                                          input_ports=[{SAMPLE:sample_input},
            #                                                        {TARGET:target_input}],
            #                                          function=LinearCombination(weights=[[-1], [1]]),
            #                                          output_ports=[ERROR_SIGNAL,
            #                                                         {NAME:MSE,
            #                                                          ASSIGN:lambda x: np.sum(x*x)/len(x)}],
            #                                          name="\'{}\' {}".format(lc.activation_output_mech.name,
            #                                                                  COMPARATOR_MECHANISM))

            objective_mechanism._role = LEARNING
            objective_mechanism._learning_role = TARGET

        try:
            lc.error_projection = objective_mechanism.input_port.path_afferents[0]
            # FIX: THIS IS TO FORCE ASSIGNMENT (SINCE IT DOESN'T SEEM TO BE
            # ASSIGNED BY TEST BELOW)
        except AttributeError:
            raise LearningAuxiliaryError("PROGRAM ERROR: problem finding "
                                          "projection to TARGET "
                                          "ObjectiveMechanism from {} when "
                                          "instantiating {}".format(
                lc.activation_output_mech.name,
                learning_projection.name))
        else:
            if not lc.error_matrix:
                raise LearningAuxiliaryError("PROGRAM ERROR: problem "
                                              "assigning error_matrix for "
                                              "projection to "
                                              "ObjectiveMechanism for {} when "
                                              "instantiating {}".format(
                    lc.activation_output_mech.name,
                    learning_projection.name))

        # INSTANTIATE LearningMechanism

    # - LearningMechanism incoming projections (by inputPort):
    #    ACTIVATION_INPUT:
    #        MappingProjection from activation_mech_input
    #    ACTIVATION_OUTPUT:
    #        MappingProjection from activation_mech_output
    #    ERROR_SIGNAL:
    #        specified in error_source argument:
    #        Note:  the error_source for LearningMechanism is set in lc.error_signal_mech:
    #            if is_target, this comes from the primary outputPort of objective_mechanism;
    #            otherwise, it comes from output_ports[ERROR_SIGNAL] of the LearningMechanism for lc.error_mech
    # - Use of AUTO_ASSIGN_MATRIX for the MappingProjections is safe, as compatibility of senders and receivers
    #    is checked in the instantiation of the learning_function

    error_source = lc.error_signal_mech

    learning_mechanism = LearningMechanism(default_variable=[activation_input,
                                                     activation_output,
                                                     error_signal],
                                           error_sources=error_source,
                                           function=learning_function,
                                           # learning_signals=[lc.activation_mech_projection],
                                           learning_signals=[learning_projection],
                                           name=LEARNING_MECHANISM + " for " + lc.activation_mech_projection.name)
    learning_mechanism.learning_type = LearningType.SUPERVISED
    learning_mechanism.learning_timing = LearningTiming.LEARNING_PHASE

    # IMPLEMENTATION NOTE:
    # ADD ARGUMENTS TO LearningMechanism FOR activation_input AND activation_output, AND THEN INSTANTIATE THE
    # MappingProjections BELOW IN CALL TO HELPER METHODS FROM LearningMechanism._instantiate_attributes_before_function
    # (FOLLOWING DESIGN OF _instantiate_error_signal_projection IN LearningMechanism):

    # Assign MappingProjection from activation_mech_input to LearningMechanism's ACTIVATION_INPUT inputPort
    lc._activation_mech_input_projection = MappingProjection(sender=lc.activation_mech_input,
                      receiver=learning_mechanism.input_ports[ACTIVATION_INPUT],
                      matrix=IDENTITY_MATRIX,
                      name = lc.activation_mech_input.owner.name + ' to ' + ACTIVATION_INPUT)

    # Assign MappingProjection from activation_mech_output to LearningMechanism's ACTIVATION_OUTPUT inputPort
    lc._activation_mech_output_projection = MappingProjection(sender=lc.activation_mech_output,
                      receiver=learning_mechanism.input_ports[ACTIVATION_OUTPUT],
                      matrix=IDENTITY_MATRIX,
                      name = lc.activation_mech_output.owner.name + ' to ' + ACTIVATION_OUTPUT)

    lc.learning_mechanism = learning_mechanism

    learning_projection._learning_components = lc


def _instantiate_error_signal_projection(sender, receiver):
    """Instantiate a MappingProjection to carry an error_signal to a LearningMechanism

    Can take as the sender an `ObjectiveMechanism` or a `LearningMechanism`.
    If the sender is an ObjectiveMechanism, uses its `primary OutputPort <OutputPort_Primary>`.
    If the sender is a LearningMechanism, uses its `ERROR_SIGNAL <LearningMechanism.output_ports>` OutputPort.
    The receiver must be a LearningMechanism; its `ERROR_SIGNAL <LearningMechanism.input_ports>` InputPort is used.
    Uses and IDENTITY_MATRIX for the MappingProjection, so requires that the sender be the same length as the receiver.

    """
    from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection

    if isinstance(sender, ObjectiveMechanism):
        sender = sender.output_ports[OUTCOME]
    elif isinstance(sender, LearningMechanism):
        sender = sender.output_ports[ERROR_SIGNAL]
    else:
        raise LearningAuxiliaryError("Sender of the error signal Projection {} "
                                     "must be either an ObjectiveMechanism or "
                                     "a LearningMechanism".format(sender))

    if isinstance(receiver, LearningMechanism):
        receiver = receiver.input_ports[ERROR_SIGNAL]
    else:
        raise LearningAuxiliaryError("Receiver of the error signal Projection "
                                     "{} must be a LearningMechanism".format(receiver))

    if len(sender.defaults.value) != len(receiver.defaults.value):
        raise LearningAuxiliaryError("The length of the OutputPort ({}) for "
                                     "the sender ({}) of the error signal "
                                     "Projection does not match the length of "
                                     "the InputPort ({}) for the receiver "
                                     "({})".format(len(sender.defaults.value),
                                                   sender.owner.name,
                                                   len(receiver.defaults.value),
                                                   receiver.owner.name))

    return MappingProjection(sender=sender,
                             receiver=receiver,
                             matrix=IDENTITY_MATRIX,
                             name=sender.owner.name + ' to ' + ERROR_SIGNAL)


@tc.typecheck
def _get_learning_mechanisms(mech:Mechanism, composition=None):
    """Return LearningMechanisms for all Projections to and from the specified Mechanism.

    If composition is specified, only LearningMechanisms for Projections to Mechanisms belonging to that composition
    (i.e., Process or System) are included in the list of LearningMechanisms returned

    Returns two lists:
        - one (aff) with the LearningMechanisms for the Mechanism's afferent projections
        - the other (eff) with the LearningMechanisms for Mechanism's efferent Projections
    """

    from psyneulink.core.components.system import System
    from psyneulink.core.components.process import Process
    aff = []
    eff = []

    if composition and not isinstance(composition, (Process, System)):
        raise LearningAuxiliaryError("composition argument for _get_learing_mechanisms ({}) must be a {} or {}".
                                     format(composition, Process.__name__, System.__name__))

    if isinstance(composition, Process):
        composition_attrib = 'processes'
    elif isinstance(composition, System):
        composition_attrib = 'systems'

    for projection in mech.path_afferents:
        if projection.has_learning_projection and (composition is None
                                                   or composition in projection.receiver.owner.processes
                                                   or composition in projection.receiver.owner.systems):

            aff.extend([learning_projection.sender.owner
                        for learning_projection in projection.parameter_ports[MATRIX].mod_afferents
                        if isinstance(learning_projection, LearningProjection)
                        and (not composition or composition in getattr(learning_projection.sender.owner,
                                                                       composition_attrib))
                        ])

    for projection in mech.efferents:
        if projection.has_learning_projection and (composition is None
                                                   or composition in projection.receiver.owner.processes
                                                   or composition in projection.receiver.owner.systems):
            eff.extend([learning_projection.sender.owner
                        for learning_projection in projection.parameter_ports[MATRIX].mod_afferents
                        if isinstance(learning_projection, LearningProjection)
                        and (not composition or composition in getattr(learning_projection.sender.owner,
                                                                       composition_attrib))
                       ])
    return aff, eff


def _assign_error_signal_projections(processing_mech:Mechanism,
                                     system,
                                     scope=None,
                                     objective_mech:tc.optional(ObjectiveMechanism)=None):
    """Assign appropriate error_signal Projections to LearningMechanisms for processing_mechanism's afferents.

    Assign an error_signal Projection to the LearningMechanism for each afferent Projection of processing_mechanism
    that is being learned, from the LearningMechanism of each of processing_mechanism's efferents that is being learned,
    unless such a projection already exists

    Scope can be a Process, System, or both (Process must be one in the System)_
    If scope is specified:
       - and it i
     [??and only for afferents and efferents that belong to the same System.]

    system argument is used to assign System to affected LearningMechanisms
    """

    # composition = None
    # if isinstance(scope, list):
    #     if len(scope)!=2:
    #         raise LearningAuxiliaryError("PROGRAM ERROR: Can only have two args: "
    #                                      "one must be System and the other a Proccess ({})".format(scope))
    #     for item in scope:
    #         if isinstance(item, Process):
    #             proc = item
    #         if isinstance(item, System):
    #             sys = item
    #         if not proc in sys.processes:
    #             raise LearningAuxiliaryError("PROGRAM ERROR: Proccess ({}) must be in System ({}) specified in scope" .
    #                                          format(proc, sys))
    # else:
    #     composition = scope
    composition = scope

    # Get all LearningMechanisms for Projection to and from sample_mech (processing_mechanism)
    afferent_learning_mechs, efferent_learning_mechs = _get_learning_mechanisms(processing_mech, scope)

    # For the LearningMechanism of each Projection to sample_mech that is being learned
    for aff_lm in afferent_learning_mechs:
        # Check that aff_lm receives in its ACTIVATION_OUTPUT InputPort the same Projection
        #    that the ObjectMechanism received
        if objective_mech and not (aff_lm.input_ports['activation_output'].path_afferents[0].sender ==
                                   objective_mech.input_ports[SAMPLE].path_afferents[0].sender):
            raise LearningAuxiliaryError("PROGRAM ERROR: The {} being assigned to replace {} ({}) receives its "
                                          "ACTIVATION_OUTPUT Projection from a source ({}) that is different than "
                                          "the source of the {}'s SAMPLE InputPort ({})."
                                         .format(LearningMechanism, objective_mech.name, aff_lm.name,
                                                  aff_lm.input_ports['activation_output'].path_afferents[0].sender,
                                                  aff_lm.name,
                                                  objective_mech.input_ports[SAMPLE].path_afferents[0].sender.name))
        # For each Projection from sample_mech that is being learned,
        #    add a Projection from its LearningMechanism ERROR_SIGNAL OutputPort
        #    to a newly created ERROR_SIGNAL InputPort on afferent_lm
        for eff_lm in efferent_learning_mechs:
            # Make sure Projection doesn't already exist
            if not any(proj.sender.owner == eff_lm for proj in aff_lm.afferents if ERROR_SIGNAL in proj.receiver.name):
                # aff_lm.add_ports(InputPort(variable=eff_lm.output_ports[ERROR_SIGNAL].value,
                #                              projections=eff_lm.output_ports[ERROR_SIGNAL],
                #                              name=ERROR_SIGNAL))
                aff_lm.add_ports(InputPort(projections=eff_lm.output_ports[ERROR_SIGNAL],
                                            name=ERROR_SIGNAL,
                                            context=Context(source=ContextFlags.METHOD)),
                                  context=Context(source=ContextFlags.METHOD))

        for projection in aff_lm.projections:
            projection._activate_for_compositions(system)

        if not aff_lm.systems:
            aff_lm._add_system(system, LEARNING)


class LearningComponents(object):
    """Gets components required to instantiate LearningMechanism and its Objective Function for a LearningProjection

    Has attributes for the following learning components relevant to a `LearningProjection`,
    each of which is found and/or validated if necessary before assignment:

    * `activation_mech_projection` (`MappingProjection`):  one being learned)
    * `activation_mech_input` (`OutputPort`):  input to Mechanism to which Projection being learned Projections
    * `activation_output_mech` (`ProcessingMechanism <ProcessingMechanism>`):  Mechanism to which projection being learned
                                                                        projects
    * `activation_mech_output` (`OutputPort`):  output of activation_output_mech
    * `activation_mech_fct` (function):  function of Mechanism to which projection being learned projects
    * `activation_derivative` (function):  derivative of function for activation_output_mech
    * `error_projection` (`MappingProjection`):  next projection in learning sequence after activation_mech_projection
    * `error_matrix` (`ParameterPort`):  parameterPort of error_projection with error_matrix
    * `error_derivative` (function):  deriviative of function of error_mech
    * `error_mech` (ProcessingMechanism):  Mechanism to which error_projection projects
    * `error_mech_output` (OutputPort):  outputPort of error_mech, that projects either to the next
                                          ProcessingMechanism in the pathway, or to an ObjectiveMechanism
    * `error_signal_mech` (LearningMechanism or ObjectiveMechanism):  Mechanism from which LearningMechanism
                                                                      gets its error_signal (ObjectiveMechanism for
                                                                      the last Mechanism in a learning sequence; next
                                                                      LearningMechanism in the sequence for all others)
    * `error_signal_mech_output` (OutputPort): outputPort of error_signal_mech, that projects to the preceeding
                                                LearningMechanism in the learning sequence (or nothing for the 1st mech)

    IMPLEMENTATION NOTE:  The helper methods in this class (that assign values to the various attributes)
                          respect membership in a process;  that is, they do not assign attribute values to
                          objects that belong to a process in which the root attribute (self.activation_output_mech)
                          belongs
    """

    def __init__(self, learning_projection, context=None):

        self._validate_learning_projection(learning_projection)
        self.learning_projection = learning_projection

        self._activation_mech_projection = None
        self._activation_output_mech = None
        self._activation_mech_input = None
        self._activation_mech_input_projection = None
        self._activation_mech_fct = None
        self._activation_derivative = None
        self._activation_mech_output = None
        self._activation_mech_output_projection = None
        self._error_projection = None
        self._error_matrix = None
        self._error_derivative = None
        self._error_mech = None
        self._error_mech_output = None
        self._error_signal_mech = None
        self._error_signal_mech_output = None
        # self._error_objective_mech = None
        # self._error_objective_mech_output = None


    def _validate_learning_projection(self, learning_projection):

        if not isinstance(learning_projection, LearningProjection):
            raise LearningAuxiliaryError("{} is not a LearningProjection".format(learning_projection.name))

    # HELPER METHODS FOR lc_components
    # ---------------------------------------------------------------------------------------------------------------
    # activation_mech_projection:  one being learned) (MappingProjection)
    @property
    def activation_mech_projection(self):
        def _get_act_proj():
            try:
                self.activation_mech_projection = self.learning_projection.receiver.owner
                return self.learning_projection.receiver.owner
            except AttributeError:
                raise LearningAuxiliaryError("activation_mech_projection not identified: learning_projection ({})"
                                              "not appear to have been assiged a receiver.".
                                             format(self.learning_projection))
        return self._activation_mech_projection or _get_act_proj()

    @activation_mech_projection.setter
    def activation_mech_projection(self, assignment):
        if isinstance(assignment, (MappingProjection)):
            self._activation_mech_projection = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to activation_mech_projection; "
                                          "it must be a MappingProjection.")


    # ---------------------------------------------------------------------------------------------------------------
    # activation_output_mech:  mechanism to which projection being learned projects (ProcessingMechanism)
    @property
    def activation_output_mech(self):
        def _get_act_mech():
            if not self.activation_mech_projection:
                return None
            try:
                self.activation_output_mech = self.activation_mech_projection.receiver.owner
                return self.activation_mech_projection.receiver.owner
            except AttributeError:
                raise LearningAuxiliaryError("activation_output_mech not identified: activation_mech_projection ({})"
                                              "not appear to have been assiged a receiver.".
                                             format(self.learning_projection))
        return self._activation_output_mech or _get_act_mech()

    @activation_output_mech.setter
    def activation_output_mech(self, assignment):
        if isinstance(assignment, (ProcessingMechanism_Base)):
            self._activation_output_mech = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to activation_output_mech; "
                                          "it must be a ProcessingMechanism.")


    # ---------------------------------------------------------------------------------------------------------------
    # activation_mech_input:  input to mechanism to which projection being learned projections (OutputPort)
    @property
    def activation_mech_input(self):
        def _get_act_input():
            if not self.activation_mech_projection:
                return None
            try:
                self.activation_mech_input = self.activation_mech_projection.sender
                return self.activation_mech_projection.sender
            except AttributeError:
                raise LearningAuxiliaryError("activation_mech_input not identified: activation_mech_projection ({})"
                                              "not appear to have been assiged a sender.".
                                             format(self.activation_mech_projection))
        return self._activation_mech_input or _get_act_input()

    @activation_mech_input.setter
    def activation_mech_input(self, assignment):
        if isinstance(assignment, (OutputPort)):
            self._activation_mech_input = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to activation_mech_input; "
                                          "it must be a OutputPort.")


    # ---------------------------------------------------------------------------------------------------------------
    # activation_mech_fct:  function of mechanism to which projection being learned projects (function)
    @property
    def activation_mech_fct(self):
        def _get_act_mech_fct():
            if not self.activation_output_mech:
                return None
            try:
                self.activation_mech_fct = self.activation_output_mech.function
                return self.activation_output_mech.function
            except AttributeError:
                raise LearningAuxiliaryError("activation_mech_fct not identified: activation_output_mech ({})"
                                              "not appear to have been assiged a Function.".
                                             format(self.learning_projection))
        return self._activation_mech_fct or _get_act_mech_fct()

    @activation_mech_fct.setter
    def activation_mech_fct(self, assignment):
        if isinstance(assignment, (Function)):
            self._activation_mech_fct = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to activation_mech_fct; "
                                          "it must be a Function.")


    # ---------------------------------------------------------------------------------------------------------------
    # activation_derivative:  derivative of function for activation_output_mech (function)
    @property
    def activation_derivative(self):
        def _get_act_deriv():
            if not self.activation_mech_fct:
                return None
            try:
                self._activation_derivative = self.activation_output_mech.function.derivative
                return self.activation_output_mech.function.derivative
            except AttributeError:
                raise LearningAuxiliaryError("activation_derivative not identified: activation_mech_fct ({})"
                                              "not appear to have a derivative defined.".
                                             format(self.learning_projection))
        return self._activation_derivative or _get_act_deriv()

    @activation_derivative.setter
    def activation_derivative(self, assignment):
        if isinstance(assignment, (function_type, method_type)):
            self._activation_derivative = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to activation_derivative; "
                                          "it must be a function or method.")


    # ---------------------------------------------------------------------------------------------------------------
    # activation_mech_output:  output of activation_output_mech (OutputPort)
    @property
    def activation_mech_output(self):
        def _get_act_sample():
            if not self.activation_output_mech:
                return None
            # If MONITOR_FOR_LEARNING specifies an outputPort, use that
            try:
                sample_port_Name = self.activation_output_mech.paramsCurrent[MONITOR_FOR_LEARNING]
                self.activation_mech_output = self.activation_output_mech.output_ports[sample_port_Name]
                if not isinstance(self.activation_mech_output, OutputPort):
                    raise LearningAuxiliaryError("The specification of the MONITOR_FOR_LEARNING parameter ({}) "
                                                  "for {} is not an outputPort".
                                                 format(self.activation_mech_output, self.learning_projection))
            except KeyError:
                # No outputPort specified so use primary outputPort
                try:
                    self.activation_mech_output = self.activation_output_mech.output_port
                except AttributeError:
                    raise LearningAuxiliaryError("activation_mech_output not identified: activation_output_mech ({})"
                                                  "not appear to have been assigned a primary outputPort.".
                                                 format(self.learning_projection))
            return self.activation_output_mech.output_port

        return self._activation_mech_output or _get_act_sample()

    @activation_mech_output.setter
    def activation_mech_output(self, assignment):
        if isinstance(assignment, (OutputPort)):
            self._activation_mech_output =assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to activation_mech_output; "
                                          "it must be a OutputPort.")


    # ---------------------------------------------------------------------------------------------------------------
    # error_matrix:  parameterPort for error_matrix (ParameterPort)
    # This must be found
    @property
    def error_matrix(self):

        # Find from error_projection
        def _get_err_matrix():
            if not self.error_projection:
                return None
            try:
                self.error_matrix = self.error_projection._parameter_ports[MATRIX]
                return self.error_projection._parameter_ports[MATRIX]
            except AttributeError:
                raise LearningAuxiliaryError("error_matrix not identified: error_projection ({})"
                                              "not not have a {} parameterPort".
                                             format(self.error_projection))
        return self._error_matrix or _get_err_matrix()

    @error_matrix.setter
    def error_matrix(self, assignment):
        if isinstance(assignment, (ParameterPort)):
            self._error_matrix = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_matrix; "
                                          "it must be a ParameterPort.")


    # ---------------------------------------------------------------------------------------------------------------
    # error_projection:  one that has the error_matrix (MappingProjection)
    @property
    def error_projection(self):
        # Find from activation_output_mech:
        def _get_error_proj():
            if not self.activation_mech_output:
                return None
            projections = self.activation_mech_output.efferents
            # error_proj must be a MappingProjection that:
            #    - projects to another Mechanism in the same Process as the one that projects to it and
            #    - has a LearningProjection to it
            error_proj = next((projection for projection in projections if
                               (isinstance(projection, MappingProjection) and
                                projection.has_learning_projection and
                                any(process in projection.receiver.owner.processes
                                    # for process in self.activation_output_mech.processes))),
                                    for process in self.activation_mech_input.owner.processes))),
                              None)
            if not error_proj:
                return None
            self.error_projection = error_proj
            return error_proj
        return self._error_projection or _get_error_proj()

    @error_projection.setter
    def error_projection(self, assignment):
        if isinstance(assignment, (MappingProjection)):
            self._error_projection = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_projection; "
                                          "it must be a MappingProjection.")


    # ---------------------------------------------------------------------------------------------------------------
    # error_mech:  mechanism to which error_projection projects (ProcessingMechanism)
    @property
    def error_mech(self):

        # Find from error_projection
        def _get_err_mech():
            if not self.error_projection:
                return None
            try:
                self.error_mech = self.error_projection.receiver.owner
            except AttributeError:
                raise LearningAuxiliaryError("error_mech not identified: error_projection ({})"
                                              "does not appear to have a receiver or owner".
                                             format(self.error_projection))
            if not isinstance(self.error_mech, ProcessingMechanism_Base):
                raise LearningAuxiliaryError("error_mech found ({}) but it does not "
                                              "appear to be a ProcessingMechanism".
                                             format(self.error_mech.name))
            return self.error_projection.receiver.owner
        return self._error_mech or _get_err_mech()

    @error_mech.setter
    def error_mech(self, assignment):
        if isinstance(assignment, (ProcessingMechanism_Base)):
            self._error_mech = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_mech; "
                                          "it must be a ProcessingMechanism.")


    # -----------------------------------------------------------------------------------------------------------
    # error_derivative:  derivative of function of error_mech (function)
    @property
    def error_derivative(self):
        # Find from error_mech:
        def _get_error_deriv():
            if not self.error_mech:
                return None
            try:
                self.error_derivative = self.error_mech.function.derivative
                return self.error_mech.function.derivative
            except AttributeError:
                raise LearningAuxiliaryError("error_derivative not identified: the function ({}) "
                                              "for error_mech ({}) does not have a derivative attribute".
                                             format(self.error_mech.function.__class__.__name__,
                                                     self.error_mech.name))
        return self._error_derivative or _get_error_deriv()

    @error_derivative.setter
    def error_derivative(self, assignment):
        if isinstance(assignment, (function_type, method_type)):
            self._error_derivative = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_derivative; "
                                          "it must be a function.")


    # ---------------------------------------------------------------------------------------------------------------
    # error_mech_output: outputPort of mechanism to which error_projection projects (OutputPort)
    @property
    def error_mech_output(self):
        # Find from error_mech
        def _get_err_mech_out():
            if not self.error_mech:
                return None
            try:
                self.error_mech_output = self.error_mech.output_port
            except AttributeError:
                raise LearningAuxiliaryError("error_mech_output not identified: error_mech ({})"
                                              "does not appear to have an outputPort".
                                             format(self.error_mech_output))
            if not isinstance(self.error_mech_output, OutputPort):
                raise LearningAuxiliaryError("error_mech_output found ({}) but it does not "
                                              "appear to be an OutputPort".
                                             format(self.error_mech_output.name))
            return self.error_mech.output_port
        return self._error_mech_output or _get_err_mech_out()

    @error_mech_output.setter
    def error_mech_output(self, assignment):
        if isinstance(assignment, (OutputPort)):
            self._error_mech_output = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_mech_output; "
                                          "it must be an OutputPort.")

    # ---------------------------------------------------------------------------------------------------------------
    # error_signal_mech:  learning mechanism for error_projection (LearningMechanism)
    @property
    def error_signal_mech(self):
        # Find from error_matrix:
        def _get_error_signal_mech():
            if not self.error_matrix:
                return None
            # search the projections to the error_matrix ParameterPort for a LearningProjection
            learning_proj = next((proj for proj in self.error_matrix.mod_afferents
                                 if isinstance(proj, LearningProjection)), None)
            # if there are none, the error_matrix might be for an error_projection to an ObjectiveMechanism
            #   (i.e., the TARGET mechanism)
            if not learning_proj:
                # if error_mech is the last in the learning sequence, then its error_matrix does not receive a
                #    LearningProjection, but its error_projection does project to an ObjectiveMechanism, so return that
                objective_mechanism = self.error_matrix.owner.receiver.owner
                if not isinstance(objective_mechanism, ObjectiveMechanism):
                    raise LearningAuxiliaryError("error_signal_mech not identified: error_matrix does not have "
                                                  "a LearningProjection and error_projection does not project to a "
                                                  "TARGET ObjectiveMechanism")
                else:
                    self.error_signal_mech = objective_mechanism
                    return self.error_signal_mech
            try:
                learning_mech = learning_proj.sender.owner
            except AttributeError:
                raise LearningAuxiliaryError("error_signal_mech not identified: "
                                              "the LearningProjection to error_matrix does not have a sender")
            if not isinstance(learning_mech, LearningMechanism):
                raise LearningAuxiliaryError("error_signal_mech not identified: "
                                              "the LearningProjection to error_matrix does not come from a "
                                              "LearningMechanism")

            # Get the current Process (the one to which the current and preceding Mechanism in the sequence belong)
            current_processes = (set(self.activation_output_mech.processes.keys()).
                                 intersection(set(self.activation_mech_input.owner.processes.keys())))

            # If the learning_mech found has already been assigned to a Process that is not the current Process,
            #    then the current mechanism (self.activation_output_mech) may:
            #    - be the TERMINAL for its process,
            #    - or it may project to the same next Mechanism as the other Process

            # if learning_mech.processes and not any(process in self.activation_mech_input.owner.processes
            #                                        for process in learning_mech.processes):
            if learning_mech.processes and not any(process in current_processes
                                                   for process in learning_mech.processes):
               # Look for an ObjectiveMechanism that projects to the current Mechanism (self.activation_output_mech)
                objective_learning_proj = next((projection for projection in
                                                self.activation_output_mech.output_port.efferents
                                                if isinstance(projection.receiver.owner, ObjectiveMechanism)), None)
                # Return ObjectiveMechanism found as error_signal_mech
                if objective_learning_proj:
                    learning_mech = objective_learning_proj.receiver.owner

                # No ObjectiveMechanism was found,
                #    so check if error_projection projects to another Mechanism in the same Process:
                #        - if it doesn't, throw an exception
                #        - if it does, assign learning_mech to any shared Processes, and return as error_signal_mech
                # elif any(process in learning_proj.receiver.owner.receiver.owner.processes
                else:
                    next_mech_processes = set(self.error_projection.receiver.owner.processes.keys())
                    shared_processes = current_processes.intersection(next_mech_processes)
                    if not shared_processes:
                        raise LearningAuxiliaryError("PROGRAM ERROR: expected to find that {} projects "
                                                      "to another Mechanism in {}".
                                                     format(self.activation_output_mech.name, current_processes))
                    for process in shared_processes:
                        if not process in learning_mech.processes:
                            learning_mech._add_process(process, LEARNING)

            self.error_signal_mech = learning_mech
            return self.error_signal_mech

        return self._error_signal_mech or _get_error_signal_mech()

    @error_signal_mech.setter
    def error_signal_mech(self, assignment):
        if (assignment is None or
                isinstance(assignment, LearningMechanism) or
                (isinstance(assignment, ObjectiveMechanism) and assignment._role is LEARNING)):
            self._error_signal_mech = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment "
                                          "to error_signal_mech; it must be a "
                                          "LearningMechanism.")

    # ---------------------------------------------------------------------------------------------------------------
    # FIX: MODIFY TO RETURN EITHER outputPort if it is an ObjectiveMechanism or
    #                              output_ports[ERROR_SIGNAL] if it it a LearningMechanism
    # error_signal_mech_output: outputPort of LearningMechanism for error_projection (OutputPort)
    @property
    def error_signal_mech_output(self):
        # Find from error_mech
        def _get_err_sig_mech_out():
            if not self.error_signal_mech:
                return None
            if isinstance(self.error_signal_mech, ObjectiveMechanism):
                try:
                    self.error_signal_mech_output = self.error_signal_mech.output_port
                except AttributeError:
                    raise LearningAuxiliaryError("error_signal_mech_output "
                                                  "not identified: "
                                                  "error_signal_mech ({}) does "
                                                  "not appear to have an "
                                                  "OutputPort".
                                                 format(self.error_signal_mech.name))
                if not isinstance(self.error_signal_mech_output, OutputPort):
                    raise LearningAuxiliaryError("error_signal_mech_output "
                                                  "found ({}) for {} but it "
                                                  "does not appear to be an "
                                                  "OutputPort".
                                                 format(self.error_signal_mech.name,
                                                         self.error_signal_mech_output.name))
            elif isinstance(self.error_signal_mech, LearningMechanism):
                try:
                    self.error_signal_mech_output = self.error_signal_mech.output_ports[ERROR_SIGNAL]
                except AttributeError:
                    raise LearningAuxiliaryError("error_signal_mech_output "
                                                  "not identified: "
                                                  "error_signal_mech ({}) does "
                                                  "not appear to have an "
                                                  "ERROR_SIGNAL outputPort".
                                                 format(self.error_signal_mech))
                if not isinstance(self.error_signal_mech_output, OutputPort):
                    raise LearningAuxiliaryError("error_signal_mech_output "
                                                  "found ({}) for {} but it "
                                                  "does not appear to be an "
                                                  "OutputPort".format(self.error_signal_mech.name,
                                                                       self.error_signal_mech_output.name))
            return self.error_signal_mech.output_port
        return self._error_signal_mech_output or _get_err_sig_mech_out()

    @error_signal_mech_output.setter
    def error_signal_mech_output(self, assignment):
        if isinstance(assignment, (OutputPort)):
            self._error_signal_mech_output = assignment
        else:
            raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment "
                                          "to error_signal_mech_output; it "
                                          "must be an OutputPort.")

    # # ---------------------------------------------------------------------------------------------------------------
    # # error_objective_mech:  TARGET objective mechanism for error_mech (ObjectiveMechanism)
    # @property
    # def error_objective_mech(self):
    #     # Find from error_matrix:
    #     def _get_obj_mech():
    #         if not self.error_matrix:
    #             return None
    #         learning_proj = next((proj for proj in self.error_matrix.path_afferents
    #                              if isinstance(proj, LearningProjection)),None)
    #         if not learning_proj:
    #             # error_matrix is for a MappingProjection that projects to the TARGET ObjectiveMechanism, so return that
    #             if isinstance(self.error_matrix.owner.receiver.owner, ObjectiveMechanism):
    #                 self.error_objective_mech = self.error_matrix.owner.receiver.owner
    #                 return self.error_matrix.owner.receiver.owner
    #             else:
    #                 raise LearningAuxiliaryError("error_objective_mech not identified: error_matrix does not have a "
    #                                               "LearningProjection and is not for a TARGET ObjectiveMechanism")
    #         try:
    #             learning_mech = learning_proj.sender.owner
    #         except AttributeError:
    #             raise LearningAuxiliaryError("error_objective_mech not identified: "
    #                                           "the LearningProjection to error_matrix does not have a sender")
    #         if not isinstance(learning_mech, LearningMechanism):
    #             raise LearningAuxiliaryError("error_objective_mech not identified: "
    #                                           "the LearningProjection to error_matrix does not come from a "
    #                                           "LearningMechanism")
    #         try:
    #             error_obj_mech = next((proj.sender.owner
    #                                    for proj in learning_mech.input_ports[ERROR_SIGNAL].path_afferents
    #                                    if isinstance(proj.sender.owner, ObjectiveMechanism)),None)
    #         except AttributeError:
    #             # return None
    #             raise LearningAuxiliaryError("error_objective_mech not identified: "
    #                                           "could not find any projections to the LearningMechanism ({})".
    #                                           format(learning_mech))
    #         # if not error_obj_mech:
    #         #     raise LearningAuxiliaryError("error_objective_mech not identified: "
    #         #                                   "the LearningMechanism ({}) does not receive a projection "
    #         #                                   "from an ObjectiveMechanism".
    #         #                                   format(learning_mech))
    #         self.error_objective_mech = error_obj_mech
    #         return error_obj_mech
    #
    #     return self._error_objective_mech or _get_obj_mech()
    #
    # @error_objective_mech.setter
    # def error_objective_mech(self, assignment):
    #     if assignment is None or isinstance(assignment, (ObjectiveMechanism)):
    #         self._error_objective_mech = assignment
    #     else:
    #         raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_objective_mech; "
    #                                       "it must be an ObjectiveMechanism.")
    #
    # # ---------------------------------------------------------------------------------------------------------------
    # # error_objective_mech_output: outputPort of ObjectiveMechanism for error_projection (ObjectiveMechanism)
    # @property
    # def error_objective_mech_output(self):
    #     # Find from error_mech
    #     def _get_err_obj_mech_out():
    #         if not self.error_objective_mech:
    #             return None
    #         try:
    #             self.error_objective_mech_output = self.error_objective_mech.outputPort
    #         except AttributeError:
    #             raise LearningAuxiliaryError("error_objective_mech_output not identified: error_objective_mech ({})"
    #                                           "does not appear to have an outputPort".
    #                                           format(self.error_objective_mech_output))
    #         if not isinstance(self.error_objective_mech_output, OutputPort):
    #             raise LearningAuxiliaryError("error_objective_mech_output found ({}) but it does not "
    #                                           "appear to be an OutputPort".
    #                                           format(self.error_objective_mech_output.name))
    #         return self.error_objective_mech.outputPort
    #     return self._error_objective_mech_output or _get_err_obj_mech_out()
    #
    # @error_objective_mech_output.setter
    # def error_objective_mech_output(self, assignment):
    #     if isinstance(assignment, (OutputPort)):
    #         self._error_objective_mech_output = assignment
    #     else:
    #         raise LearningAuxiliaryError("PROGRAM ERROR: illegal assignment to error_objective_mech_output; "
    #                                       "it must be an OutputPort.")



from . import learningauxiliary
from . import learningmechanism

from .learningauxiliary import *
from .learningmechanism import *

__all__ = list(learningmechanism.__all__)
__all__.extend(learningauxiliary.__all__)



import pytest

import doctest
import psyneulink as pnl


def test_function_docs():
    fail, total = doctest.testmod(pnl.core.components.functions.function)

    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)
